# Copyright 2023 SysEleven GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron.api import extensions
from neutron.tests.unit.api import test_extensions
from neutron.tests.unit.extensions import test_l3 as test_l3_plugin
from neutron_lib.callbacks import events
from neutron_lib.callbacks import exceptions as cb_exc
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as lib_constants
from neutron_lib import context
from neutron_lib.plugins import constants as nconstants
from neutron_lib.plugins import directory
from neutron_vpnaas.db.vpn.vpn_ext_gw_db import VPNExtGWPlugin_db
from neutron_vpnaas.tests import base
from neutron_vpnaas.tests.unit.db.vpn import test_vpn_db


OVN_VPN_PLUGIN_KLASS = "neutron_vpnaas.services.vpn.ovn_plugin.VPNOVNPlugin"


class VPNOVNPluginDbTestCase(test_l3_plugin.L3NatTestCaseMixin,
                             base.NeutronDbPluginV2TestCase):
    def setUp(self, core_plugin=None, vpnaas_plugin=OVN_VPN_PLUGIN_KLASS,
              vpnaas_provider=None):

        service_plugins = {'vpnaas_plugin': vpnaas_plugin}
        plugin_str = 'neutron.tests.unit.extensions.test_l3.TestL3NatIntPlugin'
        super().setUp(plugin_str, service_plugins=service_plugins)
        ext_mgr = extensions.PluginAwareExtensionManager.get_instance()
        self.ext_api = test_extensions.setup_extensions_middleware(ext_mgr)
        self.core_plugin = directory.get_plugin()
        self.tenant_id = 'tenant1'


class TestVPNExtGw(VPNOVNPluginDbTestCase):
    def _pre_port_delete(self, admin_context, port_id):
        registry.publish(
            resources.PORT, events.BEFORE_DELETE, self,
            payload=events.DBEventPayload(
                admin_context,
                metadata={'port_check': True},
                resource_id=port_id))

    def _pre_subnet_delete(self, admin_context, subnet_id):
        registry.publish(resources.SUBNET, events.BEFORE_DELETE, self,
                         payload=events.DBEventPayload(admin_context,
                                                       resource_id=subnet_id))

    def _pre_network_delete(self, admin_context, network_id):
        registry.publish(resources.NETWORK, events.BEFORE_DELETE, self,
                         payload=events.DBEventPayload(admin_context,
                                                       resource_id=network_id))

    def _test_prevent_vpn_port_deletion(self, device_owner, gw_key):
        plugin = directory.get_plugin(nconstants.VPN)
        with self.router() as router, \
                self.port(device_owner=device_owner) as port:
            gateway = {'gateway': {
                'router_id': router['router']['id'],
                gw_key: port['port']['id'],
                'tenant_id': self.tenant_id
            }}
            admin_context = context.get_admin_context()
            plugin.create_gateway(admin_context, gateway)
            self.assertRaises(
                cb_exc.CallbackFailure,
                self._pre_port_delete, admin_context, port['port']['id'])

    def test_prevent_vpn_port_deletion_gw_port(self):
        self._test_prevent_vpn_port_deletion(
            lib_constants.DEVICE_OWNER_VPN_ROUTER_GW, 'gw_port_id')

    def test_prevent_vpn_port_deletion_transit_port(self):
        self._test_prevent_vpn_port_deletion(
            lib_constants.DEVICE_OWNER_VPN_TRANSIT_NETWORK, 'transit_port_id')

    def test_prevent_vpn_port_deletion_other_device_owner(self):
        plugin = directory.get_plugin(nconstants.VPN)
        device_owner = lib_constants.DEVICE_OWNER_VPN_TRANSIT_NETWORK
        with self.router() as router, \
                self.port(device_owner=device_owner) as transit_port, \
                self.port(device_owner='other-device-owner') as other_port:
            gateway = {'gateway': {
                'router_id': router['router']['id'],
                'transit_port_id': transit_port['port']['id'],
                'tenant_id': self.tenant_id
            }}
            admin_context = context.get_admin_context()
            plugin.create_gateway(admin_context, gateway)
            # BEFORE_DELETE event for other_port should not raise an exception
            self._pre_port_delete(admin_context, other_port['port']['id'])

    def test_prevent_vpn_subnet_deletion(self):
        plugin = directory.get_plugin(nconstants.VPN)
        with self.router() as router, self.subnet() as subnet:
            gateway = {'gateway': {
                'router_id': router['router']['id'],
                'transit_subnet_id': subnet['subnet']['id'],
                'tenant_id': self.tenant_id
            }}
            admin_context = context.get_admin_context()
            plugin.create_gateway(admin_context, gateway)
            self.assertRaises(
                cb_exc.CallbackFailure,
                self._pre_subnet_delete, admin_context, subnet['subnet']['id'])
            # should not raise an exception for other subnet id
            self._pre_subnet_delete(admin_context, "other-id")

    def test_prevent_vpn_network_deletion(self):
        plugin = directory.get_plugin(nconstants.VPN)
        with self.router() as router, self.network() as network:
            gateway = {'gateway': {
                'router_id': router['router']['id'],
                'transit_network_id': network['network']['id'],
                'tenant_id': self.tenant_id
            }}
            admin_context = context.get_admin_context()
            plugin.create_gateway(admin_context, gateway)
            self.assertRaises(
                cb_exc.CallbackFailure,
                self._pre_network_delete, admin_context,
                network['network']['id'])
            # should not raise an exception for other network id
            self._pre_network_delete(admin_context, "other-id")


class TestVPNExtGwDB(base.NeutronDbPluginV2TestCase,
                     test_vpn_db.NeutronResourcesMixin):
    def setUp(self):
        plugin_str = 'neutron.tests.unit.extensions.test_l3.TestL3NatIntPlugin'
        super().setUp(plugin_str)

        self.core_plugin = directory.get_plugin()
        self.l3_plugin = directory.get_plugin(nconstants.L3)
        self.tenant_id = 'tenant1'
        self.context = context.get_admin_context()

    def _create_gw_port(self, router):
        port = {'port': {
            'tenant_id': self.tenant_id,
            'network_id': router['external_gateway_info']['network_id'],
            'fixed_ips': lib_constants.ATTR_NOT_SPECIFIED,
            'mac_address': lib_constants.ATTR_NOT_SPECIFIED,
            'admin_state_up': True,
            'device_id': router['id'],
            'device_owner': lib_constants.DEVICE_OWNER_VPN_ROUTER_GW,
            'name': ''
        }}
        return self.core_plugin.create_port(self.context, port)

    def test_create_gateway(self):
        private_subnet, router = self.create_basic_topology()
        gateway = {'gateway': {
            'router_id': router['id'],
            'tenant_id': self.tenant_id
        }}
        gwdb = VPNExtGWPlugin_db()
        new_gateway = gwdb.create_gateway(self.context, gateway)
        expected = {**gateway['gateway'],
                    'status': lib_constants.PENDING_CREATE}
        self.assertDictSupersetOf(expected, new_gateway)

    def test_update_gateway_with_external_port(self):
        private_subnet, router = self.create_basic_topology()
        gwdb = VPNExtGWPlugin_db()
        # create gateway
        gateway = {'gateway': {
            'router_id': router['id'],
            'tenant_id': self.tenant_id
        }}
        new_gateway = gwdb.create_gateway(self.context, gateway)

        # create external port and update gateway with the port id
        gw_port = self._create_gw_port(router)
        gateway_update = {'gateway': {
            'gw_port_id': gw_port['id']
        }}
        gwdb.update_gateway(self.context, new_gateway['id'], gateway_update)

        # check that get_vpn_gw_dict_by_router_id includes external_fixed_ips
        found_gateway = gwdb.get_vpn_gw_dict_by_router_id(self.context,
                                                          router['id'])
        self.assertIn('external_fixed_ips', found_gateway)
        expected = sorted(gw_port['fixed_ips'])
        returned = sorted(found_gateway['external_fixed_ips'])
        self.assertEqual(returned, expected)

    def test_delete_gateway(self):
        private_subnet, router = self.create_basic_topology()
        gwdb = VPNExtGWPlugin_db()
        # create gateway
        gateway = {'gateway': {
            'router_id': router['id'],
            'tenant_id': self.tenant_id
        }}
        new_gateway = gwdb.create_gateway(self.context, gateway)
        self.assertIsNotNone(new_gateway)
        deleted = gwdb.delete_gateway(self.context, new_gateway['id'])
        self.assertEqual(deleted, 1)
        deleted = gwdb.delete_gateway(self.context, new_gateway['id'])
        self.assertEqual(deleted, 0)
        found_gateway = gwdb.get_vpn_gw_dict_by_router_id(self.context,
                                                          router['id'])
        self.assertIsNone(found_gateway)



# Copyright 2015 Mirantis Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""drop_tenant_id_in_cisco_csr_identifier_map

Revision ID: 2c82e782d734
Revises: 333dfd6afaa2
Create Date: 2015-08-20 15:17:09.897944

"""

from alembic import op
import sqlalchemy as sa

from neutron.db import migration


# revision identifiers, used by Alembic.
revision = '2c82e782d734'
down_revision = '333dfd6afaa2'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.LIBERTY]


def upgrade():
    insp = sa.inspect(op.get_bind())
    if 'cisco_csr_identifier_map' not in insp.get_table_names():
        return
    op.drop_column('cisco_csr_identifier_map', 'tenant_id')



# Copyright (c) 2015 Canonical, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from unittest import mock

from neutron.agent.l3 import agent as neutron_l3_agent
from neutron.agent.l3 import legacy_router
from neutron.conf.agent.l3 import config as l3_config
from neutron.tests.functional import base
from neutron_lib import constants
from oslo_config import cfg
from oslo_utils import uuidutils

from neutron_vpnaas.services.vpn import agent as vpn_agent
from neutron_vpnaas.services.vpn.device_drivers import ipsec
from neutron_vpnaas.services.vpn.device_drivers import strongswan_ipsec
from neutron_vpnaas.tests.functional.common import test_scenario

_uuid = uuidutils.generate_uuid
FAKE_ROUTER_ID = _uuid()
FAKE_IPSEC_SITE_CONNECTION1_ID = _uuid()
FAKE_IPSEC_SITE_CONNECTION2_ID = _uuid()
FAKE_IKE_POLICY = {
    'ike_version': 'v1',
    'encryption_algorithm': 'aes-128',
    'auth_algorithm': 'sha1',
    'pfs': 'group5'
}

FAKE_IPSEC_POLICY = {
    'encryption_algorithm': 'aes-128',
    'auth_algorithm': 'sha1',
    'pfs': 'group5'
}

FAKE_VPN_SERVICE = {
    'id': _uuid(),
    'router_id': FAKE_ROUTER_ID,
    'name': 'myvpn',
    'admin_state_up': True,
    'status': constants.PENDING_CREATE,
    'external_ip': '50.0.0.4',
    'subnet': {'cidr': '10.0.0.0/24'},
    'ipsec_site_connections': [
        {'peer_cidrs': ['20.0.0.0/24',
                        '30.0.0.0/24'],
         'id': FAKE_IPSEC_SITE_CONNECTION1_ID,
         'external_ip': '50.0.0.4',
         'peer_address': '30.0.0.5',
         'peer_id': '30.0.0.5',
         'psk': 'password',
         'initiator': 'bi-directional',
         'ikepolicy': FAKE_IKE_POLICY,
         'ipsecpolicy': FAKE_IPSEC_POLICY,
         'status': constants.PENDING_CREATE},
        {'peer_cidrs': ['40.0.0.0/24',
                        '50.0.0.0/24'],
         'external_ip': '50.0.0.4',
         'peer_address': '50.0.0.5',
         'peer_id': '50.0.0.5',
         'psk': 'password',
         'id': FAKE_IPSEC_SITE_CONNECTION2_ID,
         'initiator': 'bi-directional',
         'ikepolicy': FAKE_IKE_POLICY,
         'ipsecpolicy': FAKE_IPSEC_POLICY,
         'status': constants.PENDING_CREATE}]
}

DESIRED_CONN_STATUS = {FAKE_IPSEC_SITE_CONNECTION1_ID:
                       {'status': 'DOWN',
                        'updated_pending_status': False},
                       FAKE_IPSEC_SITE_CONNECTION2_ID:
                       {'status': 'DOWN',
                        'updated_pending_status': False}}

FAKE_IKE_POLICY2 = {
    'ike_version': 'v1',
    'encryption_algorithm': 'aes-256',
    'auth_algorithm': 'sha1',
    'pfs': 'group2',
    'lifetime_value': 1800
}

FAKE_IPSEC_POLICY2 = {
    'encryption_algorithm': 'aes-256',
    'auth_algorithm': 'sha1',
    'pfs': 'group2',
    'transform_protocol': 'esp',
    'lifetime_value': 1800,
    'encapsulation_mode': 'tunnel'
}


class TestStrongSwanDeviceDriver(base.BaseSudoTestCase):

    """Test the StrongSwan reference implementation of the device driver."""

    def setUp(self):
        super().setUp()
        self.conf = cfg.CONF
        self.conf.register_opts(l3_config.OPTS)
        self.conf.register_opts(ipsec.ipsec_opts, 'ipsec')
        self.conf.register_opts(strongswan_ipsec.strongswan_opts,
                                'strongswan')
        self.conf.set_override('state_path', '/tmp')

        ri_kwargs = {'router': {'id': FAKE_ROUTER_ID},
                     'agent_conf': self.conf,
                     'interface_driver': mock.sentinel.interface_driver}
        self.router = legacy_router.LegacyRouter(router_id=FAKE_ROUTER_ID,
                                                 agent=mock.Mock(),
                                                 **ri_kwargs)
        self.router.router['distributed'] = False
        self.router_id = FAKE_VPN_SERVICE['router_id']

        looping_call_p = mock.patch(
            'oslo_service.loopingcall.FixedIntervalLoopingCall')
        looping_call_p.start()

        vpn_service = mock.Mock()
        vpn_service.conf = self.conf
        self.driver = strongswan_ipsec.StrongSwanDriver(
            vpn_service, host=mock.sentinel.host)
        self.driver.routers[FAKE_ROUTER_ID] = self.router
        self.driver.agent_rpc = mock.Mock()
        self.driver._update_nat = mock.Mock()
        self.driver.agent_rpc.get_vpn_services_on_host.return_value = [
            FAKE_VPN_SERVICE]
        self.addCleanup(self.driver.destroy_router, self.router_id)

        self.router.router_namespace.create()
        self.addCleanup(self.router.router_namespace.delete)

    def test_process_lifecycle(self):
        """
        Lifecycle test that validates that the strongswan process could be
        launched, that a connection could be successfully initiated through
        it, and then that it could be terminated and clean up after itself.
        """
        process = self.driver.ensure_process(self.router_id,
                                             FAKE_VPN_SERVICE)
        process.enable()
        self.assertTrue(process.active)
        self.assertIn(self.router_id, self.driver.processes)
        self.assertEqual(DESIRED_CONN_STATUS, process.connection_status)
        self.assertIsNotNone(process.namespace)

        conf_dir = os.path.join(self.conf.ipsec.config_base_dir,
                                self.router_id)
        self.assertTrue(os.path.exists(conf_dir))
        process.disable()
        self.assertFalse(process.active)
        self.assertFalse(process.connection_status)
        self.assertFalse(os.path.exists(conf_dir))


class TestStrongSwanScenario(test_scenario.TestIPSecBase):

    def setUp(self):
        super().setUp()
        self.conf.register_opts(strongswan_ipsec.strongswan_opts,
                                'strongswan')
        VPNAAS_STRONGSWAN_DEVICE = ('neutron_vpnaas.services.vpn.'
                                    'device_drivers.strongswan_ipsec.'
                                    'StrongSwanDriver')
        cfg.CONF.set_override('vpn_device_driver',
                              [VPNAAS_STRONGSWAN_DEVICE],
                              'vpnagent')
        self.agent = neutron_l3_agent.L3NATAgentWithStateReport('agent1',
                                                                self.conf)
        self.agent.init_host()
        self.vpn_agent = vpn_agent.L3WithVPNaaS(self.conf)
        vpn_service = mock.Mock()
        vpn_service.conf = self.conf
        self.driver = strongswan_ipsec.StrongSwanDriver(
            vpn_service, host=mock.sentinel.host)

    def _override_ikepolicy_for_site(self, site, ikepolicy):
        ipsec_connection = site.vpn_service['ipsec_site_connections'][0]
        ipsec_connection['ikepolicy'] = ikepolicy

    def _override_ipsecpolicy_for_site(self, site, ipsecpolicy):
        ipsec_connection = site.vpn_service['ipsec_site_connections'][0]
        ipsec_connection['ipsecpolicy'] = ipsecpolicy

    def _override_dpd_for_site(self, site, dpdaction, dpddelay, dpdtimeout):
        ipsec_connection = site.vpn_service['ipsec_site_connections'][0]
        ipsec_connection['dpd_action'] = dpdaction
        ipsec_connection['dpd_interval'] = dpddelay
        ipsec_connection['dpd_timeout'] = dpdtimeout

    def _override_auth_algorithm_for_site(self, site, auth):
        ipsec_connection = site.vpn_service['ipsec_site_connections'][0]
        ipsec_connection['ipsecpolicy']['auth_algorithm'] = auth
        ipsec_connection['ikepolicy']['auth_algorithm'] = auth

    def test_strongswan_connection_with_non_default_value(self):
        site1 = self.create_site(test_scenario.PUBLIC_NET[4],
                                 [self.private_nets[1]])
        site2 = self.create_site(test_scenario.PUBLIC_NET[5],
                                 [self.private_nets[2]])

        self.check_ping(site1, site2, success=False)
        self.check_ping(site2, site1, success=False)

        self.prepare_ipsec_site_connections(site1, site2)
        self._override_ikepolicy_for_site(site1, FAKE_IKE_POLICY2)
        self._override_ikepolicy_for_site(site2, FAKE_IKE_POLICY2)
        self._override_ipsecpolicy_for_site(site1, FAKE_IPSEC_POLICY2)
        self._override_ipsecpolicy_for_site(site2, FAKE_IPSEC_POLICY2)
        self._override_dpd_for_site(site1, 'hold', 60, 240)
        self._override_dpd_for_site(site2, 'hold', 60, 240)
        self.sync_to_create_ipsec_connections(site1, site2)

        self.check_ping(site1, site2)
        self.check_ping(site2, site1)

    def _test_strongswan_connection_with_auth_algo(self, auth_algo):
        site1 = self.create_site(test_scenario.PUBLIC_NET[4],
                                 [self.private_nets[1]])
        site2 = self.create_site(test_scenario.PUBLIC_NET[5],
                                 [self.private_nets[2]])

        self.check_ping(site1, site2, success=False)
        self.check_ping(site2, site1, success=False)

        self.prepare_ipsec_site_connections(site1, site2)
        self._override_auth_algorithm_for_site(site1, auth_algo)
        self._override_auth_algorithm_for_site(site2, auth_algo)
        self.sync_to_create_ipsec_connections(site1, site2)

        self.check_ping(site1, site2)
        self.check_ping(site2, site1)

    def test_strongswan_connection_with_sha256(self):
        self._test_strongswan_connection_with_auth_algo('sha256')

    def test_strongswan_connection_with_sha384(self):
        self._test_strongswan_connection_with_auth_algo('sha384')

    def test_strongswan_connection_with_sha512(self):
        self._test_strongswan_connection_with_auth_algo('sha512')

    def test_strongswan_connection_with_non_ascii_vpnservice_name(self):
        site1 = self.create_site(test_scenario.PUBLIC_NET[4],
                                 [self.private_nets[1]])
        site2 = self.create_site(test_scenario.PUBLIC_NET[5],
                                 [self.private_nets[2]])
        site1.vpn_service.update(
            {'name': test_scenario.NON_ASCII_VPNSERVICE_NAME})

        self.check_ping(site1, site2, success=False)
        self.check_ping(site2, site1, success=False)

        self.prepare_ipsec_site_connections(site1, site2)
        self.sync_to_create_ipsec_connections(site1, site2)

        self.check_ping(site1, site2)
        self.check_ping(site2, site1)

    def test_strongswan_connection_with_non_ascii_psk(self):
        site1 = self.create_site(test_scenario.PUBLIC_NET[4],
                                 [self.private_nets[1]])
        site2 = self.create_site(test_scenario.PUBLIC_NET[5],
                                 [self.private_nets[2]])

        self.check_ping(site1, site2, success=False)
        self.check_ping(site2, site1, success=False)

        self.prepare_ipsec_site_connections(site1, site2)
        self._update_ipsec_connection(site1, psk=test_scenario.NON_ASCII_PSK)
        self._update_ipsec_connection(site2, psk=test_scenario.NON_ASCII_PSK)
        self.sync_to_create_ipsec_connections(site1, site2)

        self.check_ping(site1, site2)
        self.check_ping(site2, site1)

    def test_strongswan_connection_with_wrong_non_ascii_psk(self):
        site1 = self.create_site(test_scenario.PUBLIC_NET[4],
                                 [self.private_nets[1]])
        site2 = self.create_site(test_scenario.PUBLIC_NET[5],
                                 [self.private_nets[2]])

        self.check_ping(site1, site2, success=False)
        self.check_ping(site2, site1, success=False)

        self.prepare_ipsec_site_connections(site1, site2)
        self._update_ipsec_connection(site1, psk=test_scenario.NON_ASCII_PSK)
        self._update_ipsec_connection(site2,
                                      psk=test_scenario.NON_ASCII_PSK[:-1])
        self.sync_to_create_ipsec_connections(site1, site2)

        self.check_ping(site1, site2, success=False)
        self.check_ping(site2, site1, success=False)



# Copyright 2020, SysEleven GbmH
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from neutron.api.rpc.agentnotifiers import utils as ag_utils
from neutron_lib import rpc as n_rpc
import oslo_messaging

from neutron_vpnaas.services.vpn.common import topics

# default messaging timeout is 60 sec, so 2 here is chosen to not block API
# call for more than 2 minutes
AGENT_NOTIFY_MAX_ATTEMPTS = 2


class VPNAgentNotifyAPI:
    """API for plugin to notify VPN agent."""

    def __init__(self, topic=topics.IPSEC_AGENT_TOPIC):
        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)

    def agent_updated(self, context, admin_state_up, host):
        cctxt = self.client.prepare(server=host)
        cctxt.cast(context, 'agent_updated',
                   payload={'admin_state_up': admin_state_up})

    def vpnservice_removed_from_agent(self, context, router_id, host):
        """Notify agent about removed VPN service(s) of a router."""
        cctxt = self.client.prepare(server=host)
        cctxt.cast(context, 'vpnservice_removed_from_agent',
                   router_id=router_id)

    def vpnservice_added_to_agent(self, context, router_ids, host):
        """Notify agent about added VPN service(s) of router(s)."""
        # need to use call here as we want to be sure agent received
        # notification and router will not be "lost". However using call()
        # itself is not a guarantee, calling code should handle exceptions and
        # retry
        cctxt = self.client.prepare(server=host)
        call = ag_utils.retry(cctxt.call, AGENT_NOTIFY_MAX_ATTEMPTS)
        call(context, 'vpnservice_added_to_agent', router_ids=router_ids)



#    (c) Copyright 2016 IBM Corporation, All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from neutron.api import extensions
from neutron.api.v2 import resource
from neutron.api import wsgi
from neutron import policy
from neutron_lib.api import extensions as lib_extensions
from neutron_lib.api import faults as base
from neutron_lib import exceptions
from neutron_lib.plugins import constants as plugin_const
from neutron_lib.plugins import directory
from neutron_lib import rpc as n_rpc
from oslo_log import log as logging
import webob.exc


LOG = logging.getLogger(__name__)


VPN_ROUTER = 'vpn-router'
VPN_ROUTERS = VPN_ROUTER + 's'
VPN_AGENT = 'vpn-agent'
VPN_AGENTS = VPN_AGENT + 's'


class VPNRouterSchedulerController(wsgi.Controller):
    def get_plugin(self):
        plugin = directory.get_plugin(plugin_const.VPN)
        if not plugin:
            LOG.error('No plugin for VPN registered to handle VPN '
                      'router scheduling')
            msg = 'The resource could not be found.'
            raise webob.exc.HTTPNotFound(msg)
        return plugin

    def index(self, request, **kwargs):
        plugin = self.get_plugin()
        policy.enforce(request.context,
                       "get_%s" % VPN_ROUTERS,
                       {})
        return plugin.list_routers_on_vpn_agent(
            request.context, kwargs['agent_id'])

    def create(self, request, body, **kwargs):
        plugin = self.get_plugin()
        policy.enforce(request.context,
                       "create_%s" % VPN_ROUTER,
                       {})
        agent_id = kwargs['agent_id']
        router_id = body['router_id']
        result = plugin.add_router_to_vpn_agent(request.context, agent_id,
                                                router_id)
        notify(request.context, 'vpn_agent.router.add', router_id, agent_id)
        return result

    def delete(self, request, id, **kwargs):
        plugin = self.get_plugin()
        policy.enforce(request.context,
                       "delete_%s" % VPN_ROUTER,
                       {})
        agent_id = kwargs['agent_id']
        result = plugin.remove_router_from_vpn_agent(request.context, agent_id,
                                                     id)
        notify(request.context, 'vpn_agent.router.remove', id, agent_id)
        return result


class VPNAgentsHostingRouterController(wsgi.Controller):
    def get_plugin(self):
        plugin = directory.get_plugin(plugin_const.VPN)
        if not plugin:
            LOG.error('VPN plugin not registered to handle agent scheduling')
            msg = 'The resource could not be found.'
            raise webob.exc.HTTPNotFound(msg)
        return plugin

    def index(self, request, **kwargs):
        plugin = self.get_plugin()
        policy.enforce(request.context,
                       "get_%s" % VPN_AGENTS,
                       {})
        return plugin.list_vpn_agents_hosting_router(
            request.context, kwargs['router_id'])


class Vpn_agentschedulers(lib_extensions.ExtensionDescriptor):
    """Extension class supporting VPN agent scheduler.
    """

    @classmethod
    def get_name(cls):
        return "VPN Agent Scheduler"

    @classmethod
    def get_alias(cls):
        return "vpn-agent-scheduler"

    @classmethod
    def get_description(cls):
        return "Schedule VPN services of routers among VPN agents"

    @classmethod
    def get_updated(cls):
        return "2016-08-15T10:00:00-00:00"

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        exts = []
        parent = dict(member_name="agent",
                      collection_name="agents")

        controller = resource.Resource(VPNRouterSchedulerController(),
                                       base.FAULT_MAP)
        exts.append(extensions.ResourceExtension(
            VPN_ROUTERS, controller, parent))

        parent = dict(member_name="router",
                      collection_name="routers")

        controller = resource.Resource(VPNAgentsHostingRouterController(),
                                       base.FAULT_MAP)
        exts.append(extensions.ResourceExtension(
            VPN_AGENTS, controller, parent))
        return exts

    def get_extended_resources(self, version):
        return {}


class InvalidVPNAgent(exceptions.agent.AgentNotFound):
    message = "Agent %(id)s is not a VPN Agent or has been disabled"


class RouterHostedByVPNAgent(exceptions.Conflict):
    message = ("The VPN service of router %(router_id)s has been already "
               "hosted by the VPN Agent %(agent_id)s.")


class RouterSchedulingFailed(exceptions.Conflict):
    message = ("Failed scheduling router %(router_id)s to the VPN Agent "
               "%(agent_id)s.")


class RouterReschedulingFailed(exceptions.Conflict):
    message = ("Failed rescheduling router %(router_id)s: "
               "No eligible VPN agent found.")


class VPNAgentSchedulerPluginBase(metaclass=abc.ABCMeta):
    """REST API to operate the VPN agent scheduler.

    All methods must be in an admin context.
    """

    @abc.abstractmethod
    def add_router_to_vpn_agent(self, context, id, router_id):
        pass

    @abc.abstractmethod
    def remove_router_from_vpn_agent(self, context, id, router_id):
        pass

    @abc.abstractmethod
    def list_routers_on_vpn_agent(self, context, id):
        pass

    @abc.abstractmethod
    def list_vpn_agents_hosting_router(self, context, router_id):
        pass


def notify(context, action, router_id, agent_id):
    info = {'id': agent_id, 'router_id': router_id}
    notifier = n_rpc.get_notifier('router')
    notifier.info(context, action, {'agent': info})



# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import configparser as ConfigParser
import errno
import os
import subprocess
import sys

from neutron.common import config
from neutron_lib.utils import helpers
from oslo_config import cfg
from oslo_log import log as logging
from oslo_rootwrap import wrapper

from neutron_vpnaas._i18n import _

LOG = logging.getLogger(__name__)


def setup_conf():
    cli_opts = [
        cfg.DictOpt('mount_paths',
                    required=True,
                    help=_('Dict of paths to bind-mount (source:target) '
                           'prior to launch subprocess.')),
        cfg.ListOpt(
            'cmd',
            required=True,
            help=_('Command line to execute as a subprocess '
                   'provided as comma-separated list of arguments.')),
        cfg.StrOpt('rootwrap_config', default='/etc/neutron/rootwrap.conf',
                   help=_('Rootwrap configuration file.')),
    ]
    conf = cfg.CONF
    conf.register_cli_opts(cli_opts)
    return conf


def execute(cmd):
    if not cmd:
        return
    cmd = list(map(str, cmd))
    LOG.debug("Running command: %s", cmd)
    env = os.environ.copy()
    with subprocess.Popen(cmd,
                          stdin=subprocess.PIPE,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE,
                          close_fds=True, env=env) as obj:
        _stdout, _stderr = obj.communicate()
        _stdout = helpers.safe_decode_utf8(_stdout)
        _stderr = helpers.safe_decode_utf8(_stderr)
        msg = ('Command: %(cmd)s Exit code: %(returncode)s '
            'Stdout: %(stdout)s Stderr: %(stderr)s' %
            {'cmd': cmd,
                'returncode': obj.returncode,
                'stdout': _stdout,
                'stderr': _stderr})
        LOG.debug(msg)
        obj.stdin.close()
        # Pass the output to calling process
        sys.stdout.write(msg)
        sys.stdout.flush()
        return obj.returncode


def filter_command(command, rootwrap_config):
    # Load rootwrap configuration
    try:
        rawconfig = ConfigParser.RawConfigParser()
        rawconfig.read(rootwrap_config)
        rw_config = wrapper.RootwrapConfig(rawconfig)
    except ValueError as exc:
        LOG.error('Incorrect value in %(config)s: %(exc)s',
                  {'config': rootwrap_config, 'exc': exc})
        sys.exit(errno.EINVAL)
    except ConfigParser.Error:
        LOG.error('Incorrect configuration file: %(config)s',
                  {'config': rootwrap_config})
        sys.exit(errno.EINVAL)

    # Check if command matches any of the loaded filters
    filters = wrapper.load_filters(rw_config.filters_path)
    try:
        wrapper.match_filter(filters, command, exec_dirs=rw_config.exec_dirs)
    except wrapper.FilterMatchNotExecutable as exc:
        LOG.error('Command %(command)s is not executable: '
                  '%(path)s (filter match = %(name)s)',
                  {'command': command,
                   'path': exc.match.exec_path,
                   'name': exc.match.name})
        sys.exit(errno.EINVAL)
    except wrapper.NoFilterMatched:
        LOG.error('Unauthorized command: %(cmd)s (no filter matched)',
                  {'cmd': command})
        sys.exit(errno.EPERM)


def execute_with_mount():
    config.register_common_config_options()
    conf = setup_conf()
    conf()
    config.setup_logging()
    if not conf.cmd:
        LOG.error('No command provided, exiting')
        return errno.EINVAL

    if not conf.mount_paths:
        LOG.error('No mount path provided, exiting')
        return errno.EINVAL

    # Both sudoers and rootwrap.conf will not exist in the directory /etc
    # after bind-mount, so we can't use utils.execute(conf.cmd,
    # run_as_root=True). That's why we have to check here if cmd matches
    # CommandFilter
    filter_command(conf.cmd, conf.rootwrap_config)

    # Make sure the process is running in net namespace invoked by ip
    # netns exec(/proc/[pid]/ns/net) which is since Linux 3.0,
    # as we can't check mount namespace(/proc/[pid]/ns/mnt)
    # which is since Linux 3.8. For more detail please refer the link
    # http://man7.org/linux/man-pages/man7/namespaces.7.html
    if os.path.samefile(os.path.join('/proc/1/ns/net'),
                        os.path.join('/proc', str(os.getpid()), 'ns/net')):
        LOG.error('Cannot run without netns, exiting')
        return errno.EINVAL

    for path, new_path in conf.mount_paths.items():
        if not os.path.isdir(new_path):
            # Sometimes all directories are not ready
            LOG.debug('%s is not directory', new_path)
            continue
        if os.path.isdir(path) and os.path.isabs(path):
            return_code = execute(['mount', '--bind', new_path, path])
            if return_code == 0:
                LOG.info('%(new_path)s has been '
                         'bind-mounted in %(path)s',
                         {'new_path': new_path, 'path': path})
            else:
                LOG.error('Failed to bind-mount '
                          '%(new_path)s in %(path)s',
                          {'new_path': new_path, 'path': path})
    return execute(conf.cmd)


def main():
    sys.exit(execute_with_mount())


if __name__ == "__main__":
    main()



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Endpoint group types
SUBNET_ENDPOINT = 'subnet'
CIDR_ENDPOINT = 'cidr'
VLAN_ENDPOINT = 'vlan'
NETWORK_ENDPOINT = 'network'
ROUTER_ENDPOINT = 'router'

# NOTE: Type usage...
# IPSec local endpoints - subnet, IPSec peer endpoints - cidr
# BGP VPN local endpoints - network
# Direct connect style endpoints - vlan
# IMPORTANT: The ordering of these is important, as it is used in an enum
# for the database (and migration script). Only add to this list.
VPN_SUPPORTED_ENDPOINT_TYPES = [
    SUBNET_ENDPOINT, CIDR_ENDPOINT, VLAN_ENDPOINT,
    NETWORK_ENDPOINT, ROUTER_ENDPOINT,
]

AGENT_TYPE_VPN = "VPN Agent"

OVN_AGENT_VPN_SB_CFG_KEY = 'neutron:ovn-vpnagent-sb-cfg'
OVN_AGENT_VPN_DESC_KEY = 'neutron:description-vpnagent'
OVN_AGENT_VPN_ID_KEY = 'neutron:ovn-vpnagent-id'



# Copyright 2016 MingShuang Xian/IBM
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""vpn scheduler

Revision ID: 3b739d6906cf
Revises: 5f884db48ba9
Create Date: 2016-08-15 03:32:46.124718

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3b739d6906cf'
down_revision = '5f884db48ba9'


def upgrade():
    op.create_table(
        'routervpnagentbindings',
        sa.Column('router_id', sa.String(length=36),
                  unique=True, nullable=False),
        sa.Column('vpn_agent_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['router_id'], ['routers.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('router_id', 'vpn_agent_id'),
    )



#    (c) Copyright 2015 Cisco Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Multiple local subnets

Revision ID: 28ee739a7e4b
Revises: 41b509d10b5e
Create Date: 2015-09-09 20:32:54.231765

"""

from alembic import op
import sqlalchemy as sa

from neutron.db import migration


# revision identifiers, used by Alembic.
revision = '28ee739a7e4b'
down_revision = '41b509d10b5e'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.MITAKA]


def upgrade():
    op.add_column('ipsec_site_connections',
                  sa.Column('local_ep_group_id',
                            sa.String(length=36),
                            nullable=True))
    op.add_column('ipsec_site_connections',
                  sa.Column('peer_ep_group_id',
                            sa.String(length=36),
                            nullable=True))
    op.create_foreign_key(constraint_name=None,
                          source_table='ipsec_site_connections',
                          referent_table='vpn_endpoint_groups',
                          local_cols=['local_ep_group_id'],
                          remote_cols=['id'])
    op.create_foreign_key(constraint_name=None,
                          source_table='ipsec_site_connections',
                          referent_table='vpn_endpoint_groups',
                          local_cols=['peer_ep_group_id'],
                          remote_cols=['id'])
    op.alter_column('vpnservices', 'subnet_id',
                    existing_type=sa.String(length=36), nullable=True)



# Copyright 2017 Red Hat, Inc.
# Copyright 2023 SysEleven GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf as config
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import impl_idl_ovn
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import ovsdb_monitor
from oslo_log import log as logging
from ovs.db import idl
from ovsdbapp.backend.ovs_idl import connection
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp.schema.open_vswitch import impl_idl as idl_ovs
import tenacity

LOG = logging.getLogger(__name__)


class VPNAgentOvnSbIdl(ovsdb_monitor.OvnIdl):

    SCHEMA = 'OVN_Southbound'

    def __init__(self, chassis=None, events=None, tables=None):
        connection_string = config.get_ovn_sb_connection()
        ovsdb_monitor._check_and_set_ssl_files(self.SCHEMA)
        helper = self._get_ovsdb_helper(connection_string)
        if tables is None:
            tables = ('Chassis', 'SB_Global')
        for table in tables:
            helper.register_table(table)
        try:
            super().__init__(
                None, connection_string, helper, leader_only=False)
        except TypeError:
            # TODO(bpetermann) We can remove this when we require ovs>=2.12.0
            super().__init__(None, connection_string, helper)
        if chassis:
            table = ('Chassis_Private' if 'Chassis_Private' in tables
                     else 'Chassis')
            self.set_table_condition(table, [['name', '==', chassis]])
        if events:
            self.notify_handler.watch_events(events)

    @tenacity.retry(
        wait=tenacity.wait_exponential(max=180),
        reraise=True)
    def _get_ovsdb_helper(self, connection_string):
        return idlutils.get_schema_helper(connection_string, self.SCHEMA)

    def start(self):
        conn = connection.Connection(
            self, timeout=config.get_ovn_ovsdb_timeout())
        return impl_idl_ovn.OvsdbSbOvnIdl(conn)


class VPNAgentOvsIdl:

    def start(self):
        connection_string = config.cfg.CONF.ovs.ovsdb_connection
        helper = idlutils.get_schema_helper(connection_string,
                                            'Open_vSwitch')
        tables = ('Open_vSwitch', 'Bridge', 'Port', 'Interface')
        for table in tables:
            helper.register_table(table)
        ovs_idl = idl.Idl(
            connection_string, helper,
            probe_interval=config.get_ovn_ovsdb_probe_interval())
        conn = connection.Connection(
            ovs_idl, timeout=config.cfg.CONF.ovs.ovsdb_connection_timeout)
        return idl_ovs.OvsdbIdl(conn)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from neutron_vpnaas.policies import endpoint_group
from neutron_vpnaas.policies import ike_policy
from neutron_vpnaas.policies import ipsec_policy
from neutron_vpnaas.policies import ipsec_site_connection
from neutron_vpnaas.policies import vpnservice


def list_rules():
    return itertools.chain(
        endpoint_group.list_rules(),
        ike_policy.list_rules(),
        ipsec_policy.list_rules(),
        ipsec_site_connection.list_rules(),
        vpnservice.list_rules(),
    )



# Copyright 2018, Fujitsu Vietnam Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import rpc as n_rpc

from neutron_vpnaas.services.vpn.service_drivers import base_ipsec
from neutron_vpnaas.services.vpn.service_drivers import ipsec_validator


IPSEC = 'ipsec'
BASE_IPSEC_VERSION = '1.0'


class DummyIPsecVPNDriver(base_ipsec.BaseIPsecVPNDriver):
    """Dummy VPN Service Driver class for IPsec."""

    def __init__(self, service_plugin):
        super().__init__(
            service_plugin,
            ipsec_validator.IpsecVpnValidator(self))

    def create_rpc_conn(self):
        self.endpoints = [base_ipsec.IPsecVpnDriverCallBack(self)]
        self.conn = n_rpc.Connection()
        self.conn.create_consumer(
            'dummy_ipsec_driver', self.endpoints, fanout=False)
        self.conn.consume_in_threads()
        self.agent_rpc = base_ipsec.IPsecVpnAgentApi(
            'dummy_ipsec_agent', BASE_IPSEC_VERSION, self)



# Copyright 2023 SysEleven GmbH.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from neutron.agent.linux import ip_lib
from neutron.conf.agent import common as agent_config
from neutron.conf import common as common_config
from oslo_config import cfg
from oslo_utils import uuidutils

from neutron_vpnaas.services.vpn.device_drivers import ovn_ipsec
from neutron_vpnaas.tests import base
from neutron_vpnaas.tests.unit.services.vpn.device_drivers import test_ipsec

_uuid = uuidutils.generate_uuid


FAKE_PROCESS_ID = "c5b52e50-e678-491e-98dd-34e2676a6f81"
FAKE_NAMESPACE_NAME = "qvpn-c5b52e50-e678-491e-98dd-34e2676a6f81"

FAKE_GW_PORT_ID = "e95d89fb-1723-4865-876f-a1c8efed4b55"
FAKE_GW_PORT_INTERFACE_NAME = "vge95d89fb-172"
FAKE_GW_PORT_IP_ADDRESS = "20.20.20.20"
FAKE_GW_PORT_MAC_ADDRESS = "11:22:33:44:55:66"
FAKE_GW_PORT_SUBNET_ID = _uuid()
FAKE_GW_PORT_SUBNET_INFO = {
    'id': FAKE_GW_PORT_SUBNET_ID,
    'cidr': '20.20.20.0/24',
    'ip_version': 4
}
FAKE_GW_PORT = {
    'id': FAKE_GW_PORT_ID,
    'mac_address': FAKE_GW_PORT_MAC_ADDRESS,
    'fixed_ips': [{
        'ip_address': FAKE_GW_PORT_IP_ADDRESS,
        'subnet_id': FAKE_GW_PORT_SUBNET_ID
    }]
}
FAKE_TRANSIT_PORT_ID = "0eb4bdb3-fe2e-4724-bb04-f84b6a5974f8"
FAKE_TRANSIT_PORT_INTERFACE_NAME = "vr0eb4bdb3-fe2"
FAKE_TRANSIT_PORT_MAC_ADDRESS = "22:33:44:55:66:77"
FAKE_TRANSIT_PORT_IP_ADDRESS = "169.254.0.2"
FAKE_TRANSIT_PORT_SUBNET_ID = _uuid()
FAKE_TRANSIT_PORT = {
    'id': FAKE_TRANSIT_PORT_ID,
    'mac_address': FAKE_TRANSIT_PORT_MAC_ADDRESS,
    'fixed_ips': [{
        'ip_address': FAKE_TRANSIT_PORT_IP_ADDRESS,
        'subnet_id': FAKE_TRANSIT_PORT_SUBNET_ID
    }]
}


def fake_interface_driver(*args, **kwargs):
    driver = mock.Mock()
    driver.DEV_NAME_LEN = 14
    return driver


class TestDeviceManager(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.conf = cfg.CONF
        self.conf.register_opts(common_config.core_opts)
        self.conf.register_opts(agent_config.INTERFACE_DRIVER_OPTS)
        self.conf.set_override(
            'interface_driver',
            'neutron_vpnaas.tests.unit.services.vpn.device_drivers'
            '.test_ovn_ipsec.fake_interface_driver')
        self.host = "some-hostname"
        self.plugin = mock.Mock()
        self.plugin.get_subnet_info.return_value = FAKE_GW_PORT_SUBNET_INFO
        self.context = mock.Mock()

    def test_names(self):
        mgr = ovn_ipsec.DeviceManager(self.conf, self.host,
                                      self.plugin, self.context)
        port = {'id': "0df5beb8-4794-4217-acde-e6ce4875a59f"}
        name = mgr.get_interface_name(port, "internal")
        self.assertEqual(name, "vr0df5beb8-479")

        name = mgr.get_interface_name(port, "external")
        self.assertEqual(name, "vg0df5beb8-479")

        name = mgr.get_namespace_name("0df5beb8-4794-4217-acde-e6ce4875a59f")
        self.assertEqual(name, "qvpn-0df5beb8-4794-4217-acde-e6ce4875a59f")

    def test_setup_external(self):
        ext_net_id = _uuid()
        network_details = {
            'gw_port': FAKE_GW_PORT,
            'external_network': {
                'id': ext_net_id
            }
        }

        mgr = ovn_ipsec.DeviceManager(self.conf, self.host,
                                      self.plugin, self.context)

        with mock.patch.object(ip_lib, 'ensure_device_is_ready') as dev_ready:
            with mock.patch.object(mgr, 'set_default_route') as set_def_route:
                dev_ready.return_value = False

                mgr.setup_external(FAKE_PROCESS_ID, network_details)

                dev_ready.assert_called_once()
                self.plugin.get_subnet_info.assert_called_once_with(
                    FAKE_GW_PORT_SUBNET_ID
                )
                set_def_route.assert_called_once_with(
                    FAKE_NAMESPACE_NAME,
                    FAKE_GW_PORT_SUBNET_INFO,
                    FAKE_GW_PORT_INTERFACE_NAME
                )
                mgr.driver.init_l3.assert_called_once()
                mgr.driver.plug.assert_called_once()

    def test_setup_internal(self):
        network_details = {'transit_port': FAKE_TRANSIT_PORT}

        mgr = ovn_ipsec.DeviceManager(self.conf, self.host,
                                      self.plugin, self.context)

        with mock.patch.object(ip_lib, 'ensure_device_is_ready') as dev_ready:
            dev_ready.return_value = False

            mgr.setup_internal(FAKE_PROCESS_ID, network_details)

            dev_ready.assert_called_once()
            mgr.driver.init_l3.assert_called_once()
            mgr.driver.plug.assert_called_once()

    def test_list_routes(self):
        mgr = ovn_ipsec.DeviceManager(self.conf, self.host,
                                      self.plugin, self.context)
        mock_ipdev = mock.Mock()
        routes = [
            {'cidr': '192.168.111.0/24', 'via': FAKE_TRANSIT_PORT_IP_ADDRESS}
        ]
        with mock.patch.object(ip_lib, 'IPDevice') as ipdev:
            ipdev.return_value = mock_ipdev
            mock_ipdev.route.list_routes.return_value = routes
            returned = mgr.list_routes(FAKE_NAMESPACE_NAME)
            self.assertEqual(returned, routes)

    def test_del_static_routes(self):
        mgr = ovn_ipsec.DeviceManager(self.conf, self.host, self.plugin,
                                      self.context)
        mock_ipdev = mock.Mock()
        routes = [
            {'cidr': '192.168.111.0/24', 'via': FAKE_TRANSIT_PORT_IP_ADDRESS},
            {'cidr': '192.168.112.0/24', 'via': FAKE_TRANSIT_PORT_IP_ADDRESS}
        ]
        with mock.patch.object(ip_lib, 'IPDevice') as ipdev:
            ipdev.return_value = mock_ipdev
            mock_ipdev.route.list_routes.return_value = routes

            mgr.del_static_routes(FAKE_NAMESPACE_NAME)

            mock_ipdev.route.delete_route.assert_has_calls([
                mock.call(routes[0]['cidr'], via=FAKE_TRANSIT_PORT_IP_ADDRESS),
                mock.call(routes[1]['cidr'], via=FAKE_TRANSIT_PORT_IP_ADDRESS),
            ], any_order=True)


class TestOvnStrongSwanDriver(test_ipsec.IPSecDeviceLegacy):

    def setUp(self, driver=ovn_ipsec.OvnStrongSwanDriver,
              ipsec_process=ovn_ipsec.OvnStrongSwanProcess):
        conf = cfg.CONF
        conf.register_opts(common_config.core_opts)
        conf.register_opts(agent_config.INTERFACE_DRIVER_OPTS)
        conf.set_override(
            'interface_driver',
            'neutron_vpnaas.tests.unit.services.vpn.device_drivers'
            '.test_ovn_ipsec.fake_interface_driver')

        super().setUp(driver, ipsec_process)
        self.driver.nsmgr = mock.Mock()
        self.driver.nsmgr.exists.return_value = False
        self.driver.devmgr = mock.Mock()
        self.driver.devmgr.get_namespace_name.return_value = \
            FAKE_NAMESPACE_NAME
        self.driver.devmgr.list_routes.return_value = []
        self.driver.devmgr.get_existing_process_ids.return_value = []
        self.driver.agent_rpc.get_vpn_transit_network_details.return_value = {
            'transit_gateway_ip': '192.168.1.1',
        }

    def test_ensure_nat_rules(self):
        """Not applicable for OvnIPsecDriver"""
        pass

    def test_get_namespace_for_router(self):
        """Different for OvnIPsecDriver"""
        namespace = self.driver.get_namespace(FAKE_PROCESS_ID)
        self.assertEqual(FAKE_NAMESPACE_NAME, namespace)

    def test_fail_getting_namespace_for_unknown_router(self):
        """Not applicable for OvnIPsecDriver"""
        pass

    def test_create_router(self):
        """Not applicable for OvnIPsecDriver"""
        pass

    def test_destroy_router(self):
        """Not applicable for OvnIPsecDriver"""
        pass

    def test_ensure_nat_rules_with_multiple_local_subnets(self):
        """Not applicable for OvnIPsecDriver"""
        pass

    def _test_ensure_nat_rules(self):
        """Not applicable for OvnIPsecDriver"""
        pass

    def test_stale_cleanup(self):
        process = self.fake_ensure_process(FAKE_PROCESS_ID)

        self.driver.devmgr.get_existing_process_ids.return_value = [
            FAKE_PROCESS_ID]

        self.driver.agent_rpc.get_vpn_services_on_host.return_value = []
        context = mock.Mock()
        with mock.patch.object(self.driver, 'ensure_process') as ensure:
            ensure.return_value = process
            self.driver.sync(context, [])
            process.disable.assert_called()


class TestOvnOpenSwanDriver(TestOvnStrongSwanDriver):
    def setUp(self):
        super().setUp(driver=ovn_ipsec.OvnOpenSwanDriver,
                      ipsec_process=ovn_ipsec.OvnOpenSwanProcess)


class TestOvnLibreSwanDriver(TestOvnStrongSwanDriver):
    def setUp(self):
        super().setUp(driver=ovn_ipsec.OvnLibreSwanDriver,
                      ipsec_process=ovn_ipsec.OvnLibreSwanProcess)



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Populate VPN service table fields

Revision ID: 333dfd6afaa2
Revises: 56893333aa52
Create Date: 2015-07-27 16:43:59.123456

"""

from alembic import op
import netaddr
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '333dfd6afaa2'
down_revision = '56893333aa52'
depends_on = '24f28869838b'

VPNService = sa.Table('vpnservices', sa.MetaData(),
                      sa.Column('router_id', sa.String(36), nullable=False),
                      sa.Column('external_v4_ip', sa.String(16)),
                      sa.Column('external_v6_ip', sa.String(64)),
                      sa.Column('id', sa.String(36), nullable=False,
                                primary_key=True))
Router = sa.Table('routers', sa.MetaData(),
                  sa.Column('gw_port_id', sa.String(36)),
                  sa.Column('id', sa.String(36), nullable=False,
                            primary_key=True))
Port = sa.Table('ports', sa.MetaData(),
                sa.Column('id', sa.String(36), nullable=False,
                          primary_key=True))
IPAllocation = sa.Table('ipallocations', sa.MetaData(),
                        sa.Column('ip_address', sa.String(64),
                                  nullable=False, primary_key=True),
                        sa.Column('port_id', sa.String(36)))


def _migrate_external_ips(engine):
    """Use router external IPs to populate external_v*_ip entries.

    For each service, look through the associated router's
    gw_port['fixed_ips'] list and store any IPv4 and/or IPv6
    addresses into the new fields. If there are multiple
    addresses for an IP version, then only the first one will
    be stored (the same as the reference driver does).
    """
    insp = sa.inspect(engine)
    if 'cisco_csr_identifier_map' not in insp.get_table_names():
        return
    session = sa.orm.Session(bind=engine.connect())
    services = session.query(VPNService).all()
    for service in services:
        addresses = session.query(IPAllocation.c.ip_address).filter(
            service.router_id == Router.c.id,
            Router.c.gw_port_id == Port.c.id,
            Port.c.id == IPAllocation.c.port_id).all()
        have_version = []
        for address in addresses:
            version = netaddr.IPAddress(address[0]).version
            if version in have_version:
                continue
            have_version.append(version)
            update = {'external_v%s_ip' % version: address[0]}
            op.execute(VPNService.update().where(
                VPNService.c.id == service.id).values(update))
    session.commit()


def upgrade():
    # Use the router to populate the fields
    for_engine = op.get_bind()
    _migrate_external_ips(for_engine)



# Copyright (c) 2016 Yi Jing Zhu, IBM.
# Copyright (c) 2023 SysEleven GmbH
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from neutron.agent.common import utils as agent_common_utils
from neutron.agent.linux import ip_lib
from neutron_lib import constants as lib_constants
from neutron_lib import context as nctx
from oslo_concurrency import lockutils
from oslo_log import log as logging

from neutron_vpnaas.services.vpn.common import topics
from neutron_vpnaas.services.vpn.device_drivers import ipsec
from neutron_vpnaas.services.vpn.device_drivers import libreswan_ipsec
from neutron_vpnaas.services.vpn.device_drivers import strongswan_ipsec

PORT_PREFIX_INTERNAL = 'vr'
PORT_PREFIX_EXTERNAL = 'vg'
PORT_PREFIXES = {
    'internal': PORT_PREFIX_INTERNAL,
    'external': PORT_PREFIX_EXTERNAL,
}

LOG = logging.getLogger(__name__)


class DeviceManager:
    """Device Manager for ports in qvpn-xx namespace.
    It is a veth pair, one side in qvpn and the other
    side is attached to ovs.
    """

    OVN_NS_PREFIX = "qvpn-"

    def __init__(self, conf, host, plugin, context):
        self.conf = conf
        self.host = host
        self.plugin = plugin
        self.context = context
        self.driver = agent_common_utils.load_interface_driver(conf)

    def get_interface_name(self, port, ptype):
        suffix = port['id']
        return (PORT_PREFIXES[ptype] + suffix)[:self.driver.DEV_NAME_LEN]

    def get_namespace_name(self, process_id):
        return self.OVN_NS_PREFIX + process_id

    def get_existing_process_ids(self):
        """Return the process IDs derived from the existing VPN namespaces."""
        return [ns[len(self.OVN_NS_PREFIX):]
                for ns in ip_lib.list_network_namespaces()
                if ns.startswith(self.OVN_NS_PREFIX)]

    def set_default_route(self, namespace, subnet, device_name):
        device = ip_lib.IPDevice(device_name, namespace=namespace)
        gateway = device.route.get_gateway(ip_version=subnet['ip_version'])
        if gateway:
            gateway = gateway.get('gateway')
        new_gateway = subnet['gateway_ip']
        if gateway == new_gateway:
            return
        device.route.add_gateway(subnet['gateway_ip'])

    def add_routes(self, namespace, cidrs, via):
        device = ip_lib.IPDevice(None, namespace=namespace)
        for cidr in cidrs:
            device.route.add_route(cidr, via=via, metric=100, proto='static')

    def delete_routes(self, namespace, cidrs, via):
        device = ip_lib.IPDevice(None, namespace=namespace)
        for cidr in cidrs:
            device.route.delete_route(cidr, via=via, metric=100,
                                      proto='static')

    def list_routes(self, namespace, via=None):
        device = ip_lib.IPDevice(None, namespace=namespace)
        return device.route.list_routes(
            lib_constants.IP_VERSION_4, proto='static', via=via)

    def del_static_routes(self, namespace):
        device = ip_lib.IPDevice(None, namespace=namespace)
        routes = device.route.list_routes(
            lib_constants.IP_VERSION_4, proto='static')

        for r in routes:
            device.route.delete_route(r['cidr'], via=r['via'])

    def _del_port(self, process_id, ptype):
        namespace = self.get_namespace_name(process_id)
        prefix = PORT_PREFIXES[ptype]
        device = ip_lib.IPDevice(None, namespace=namespace)
        ports = device.addr.list()
        for p in ports:
            if not p['name'].startswith(prefix):
                continue
            interface_name = p['name']
            self.driver.unplug(interface_name, namespace=namespace)

    def del_internal_port(self, process_id):
        self._del_port(process_id, 'internal')

    def del_external_port(self, process_id):
        self._del_port(process_id, 'external')

    def setup_external(self, process_id, network_details):
        network = network_details["external_network"]
        vpn_port = network_details['gw_port']
        ns_name = self.get_namespace_name(process_id)
        interface_name = self.get_interface_name(vpn_port, 'external')

        if not ip_lib.ensure_device_is_ready(interface_name,
                                             namespace=ns_name):
            try:
                self.driver.plug(network['id'],
                                 vpn_port['id'],
                                 interface_name,
                                 vpn_port['mac_address'],
                                 namespace=ns_name,
                                 mtu=network.get('mtu'),
                                 prefix=PORT_PREFIX_EXTERNAL)
            except Exception:
                LOG.exception('plug external port %s failed', vpn_port)
                return None

        ip_cidrs = []
        subnets = []
        for fixed_ip in vpn_port['fixed_ips']:
            subnet_id = fixed_ip['subnet_id']
            subnet = self.plugin.get_subnet_info(subnet_id)
            net = netaddr.IPNetwork(subnet['cidr'])
            ip_cidr = '{}/{}'.format(fixed_ip['ip_address'], net.prefixlen)
            ip_cidrs.append(ip_cidr)
            subnets.append(subnet)
        self.driver.init_l3(interface_name, ip_cidrs,
                            namespace=ns_name)
        for subnet in subnets:
            self.set_default_route(ns_name, subnet, interface_name)
        return interface_name

    def setup_internal(self, process_id, network_details):
        vpn_port = network_details["transit_port"]
        ns_name = self.get_namespace_name(process_id)
        interface_name = self.get_interface_name(vpn_port, 'internal')

        if not ip_lib.ensure_device_is_ready(interface_name,
                                             namespace=ns_name):
            try:
                self.driver.plug('',
                                 vpn_port['id'],
                                 interface_name,
                                 vpn_port['mac_address'],
                                 namespace=ns_name,
                                 prefix=PORT_PREFIX_INTERNAL)
            except Exception:
                LOG.exception('plug internal port %s failed', vpn_port['id'])
                return None

        ip_cidrs = []
        for fixed_ip in vpn_port['fixed_ips']:
            ip_cidr = '{}/{}'.format(fixed_ip['ip_address'], 28)
            ip_cidrs.append(ip_cidr)
        self.driver.init_l3(interface_name, ip_cidrs,
                            namespace=ns_name)
        return interface_name


class NamespaceManager:
    def __init__(self, use_ipv6=False):
        self.ip_wrapper_root = ip_lib.IPWrapper()
        self.use_ipv6 = use_ipv6

    def exists(self, name):
        return ip_lib.network_namespace_exists(name)

    def create(self, name):
        ip_wrapper = self.ip_wrapper_root.ensure_namespace(name)
        cmd = ['sysctl', '-w', 'net.ipv4.ip_forward=1']
        ip_wrapper.netns.execute(cmd)
        if self.use_ipv6:
            cmd = ['sysctl', '-w', 'net.ipv6.conf.all.forwarding=1']
            ip_wrapper.netns.execute(cmd)

    def delete(self, name):
        try:
            self.ip_wrapper_root.netns.delete(name)
        except RuntimeError:
            msg = 'Failed trying to delete namespace: %s'
            LOG.exception(msg, name)


class OvnOpenSwanProcess(ipsec.OpenSwanProcess):
    pass


class OvnStrongSwanProcess(strongswan_ipsec.StrongSwanProcess):
    pass


class OvnLibreSwanProcess(libreswan_ipsec.LibreSwanProcess):
    pass


class IPsecOvnDriverApi(ipsec.IPsecVpnDriverApi):
    def __init__(self, topic):
        super().__init__(topic)
        self.admin_ctx = nctx.get_admin_context_without_session()

    def get_vpn_transit_network_details(self, router_id):
        cctxt = self.client.prepare()
        return cctxt.call(self.admin_ctx, 'get_vpn_transit_network_details',
                          router_id=router_id)

    def get_subnet_info(self, subnet_id):
        cctxt = self.client.prepare()
        return cctxt.call(self.admin_ctx, 'get_subnet_info',
                          subnet_id=subnet_id)


class OvnIPsecDriver(ipsec.IPsecDriver):

    def __init__(self, vpn_service, host):
        self.nsmgr = NamespaceManager()
        super().__init__(vpn_service, host)
        self.agent_rpc = IPsecOvnDriverApi(topics.IPSEC_DRIVER_TOPIC)
        self.devmgr = DeviceManager(self.conf, self.host,
                                    self.agent_rpc, self.context)

    get_router_based_iptables_manager = None

    def get_namespace(self, router_id):
        """Get namespace for VPN services of router.

        :router_id: router_id
        :returns: namespace string.
        """
        return self.devmgr.get_namespace_name(router_id)

    def _cleanup_namespace(self, router_id):
        ns_name = self.devmgr.get_namespace_name(router_id)
        if not self.nsmgr.exists(ns_name):
            return

        self.devmgr.del_internal_port(router_id)
        self.devmgr.del_external_port(router_id)
        self.nsmgr.delete(ns_name)

    def _ensure_namespace(self, router_id, network_details):
        ns_name = self.get_namespace(router_id)
        if not self.nsmgr.exists(ns_name):
            self.nsmgr.create(ns_name)

        # set up vpn external port on provider net
        self.devmgr.setup_external(router_id, network_details)

        # set up vpn internal port on transit net
        self.devmgr.setup_internal(router_id, network_details)

        return ns_name

    def destroy_process(self, process_id):
        LOG.info('process %s is destroyed', process_id)
        namespace = self.devmgr.get_namespace_name(process_id)

        # If the namespace exists but the process_id is not in the table
        # there may be an active swan process from a previous run of the agent
        # which does not have a process object in memory.
        # To be able to clean it up we need to create a dummy process object
        # here (without a vpnservice), so that destroy_process will stop
        # the swan.
        if self.nsmgr.exists(namespace) and process_id not in self.processes:
            self.ensure_process(process_id)
        super().destroy_process(process_id)
        self._cleanup_namespace(process_id)

    def create_router(self, router):
        pass

    def destroy_router(self, process_id):
        pass

    def _update_nat(self, vpnservice, func):
        pass

    def _update_route(self, vpnservice, network_details):
        router_id = vpnservice['router_id']
        gateway_ip = network_details['transit_gateway_ip']
        namespace = self.devmgr.get_namespace_name(router_id)

        old_local_cidrs = set()
        for route in self.devmgr.list_routes(namespace, via=gateway_ip):
            old_local_cidrs.add(route['cidr'])

        new_local_cidrs = set()
        for ipsec_site_conn in vpnservice['ipsec_site_connections']:
            new_local_cidrs.update(ipsec_site_conn['local_cidrs'])

        self.devmgr.delete_routes(namespace,
                                  old_local_cidrs - new_local_cidrs,
                                  gateway_ip)
        self.devmgr.add_routes(namespace,
                               new_local_cidrs - old_local_cidrs,
                               gateway_ip)

    def _sync_vpn_processes(self, vpnservices, sync_router_ids):
        # Ensure the ipsec process is enabled only for
        # - the vpn services which are not yet in self.processes
        # - vpn services whose router id is in 'sync_router_ids'
        for vpnservice in vpnservices:
            router_id = vpnservice['router_id']
            if router_id not in self.processes or router_id in sync_router_ids:
                net_details = self.agent_rpc.get_vpn_transit_network_details(
                    router_id)

                self._ensure_namespace(router_id, net_details)
                self._update_route(vpnservice, net_details)
                process = self.ensure_process(router_id, vpnservice=vpnservice)
                process.update()

    def _cleanup_stale_vpn_processes(self, vpn_router_ids):
        super()._cleanup_stale_vpn_processes(vpn_router_ids)
        # Look for additional namespaces on this node that we don't know
        # and that should be deleted
        for router_id in self.devmgr.get_existing_process_ids():
            if router_id not in vpn_router_ids:
                self.destroy_process(router_id)

    @lockutils.synchronized('vpn-agent', 'neutron-')
    def vpnservice_removed_from_agent(self, context, router_id):
        # must run under the same lock as sync()
        self.destroy_process(router_id)

    def vpnservice_added_to_agent(self, context, router_ids):
        routers = [{'id': router_id} for router_id in router_ids]
        self.sync(context, routers)


class OvnStrongSwanDriver(OvnIPsecDriver):
    def create_process(self, process_id, vpnservice, namespace):
        return OvnStrongSwanProcess(
            self.conf,
            process_id,
            vpnservice,
            namespace)


class OvnOpenSwanDriver(OvnIPsecDriver):
    def create_process(self, process_id, vpnservice, namespace):
        return OvnOpenSwanProcess(
            self.conf,
            process_id,
            vpnservice,
            namespace)


class OvnLibreSwanDriver(OvnIPsecDriver):
    def create_process(self, process_id, vpnservice, namespace):
        return OvnLibreSwanProcess(
            self.conf,
            process_id,
            vpnservice,
            namespace)



# Copyright 2015, Nachi Ueno, NTT I3, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import abc

import netaddr
import oslo_messaging

from neutron.db.models import l3agent
from neutron.db.models import servicetype
from neutron_lib import constants as lib_constants
from neutron_lib.db import api as db_api
from neutron_lib.plugins import directory

from neutron_vpnaas.db.vpn import vpn_models
from neutron_vpnaas.services.vpn import service_drivers


IPSEC = 'ipsec'
BASE_IPSEC_VERSION = '1.0'


class IPsecVpnDriverCallBack:
    """Callback for IPSecVpnDriver rpc."""

    # history
    #   1.0 Initial version

    target = oslo_messaging.Target(version=BASE_IPSEC_VERSION)

    def __init__(self, driver):
        super().__init__()
        self.driver = driver

    def _get_agent_hosting_vpn_services(self, context, host):
        plugin = directory.get_plugin()
        agent = plugin._get_agent_by_type_and_host(
            context, lib_constants.AGENT_TYPE_L3, host)
        agent_conf = plugin.get_configuration_dict(agent)
        # Retrieve the agent_mode to check if this is the
        # right agent to deploy the vpn service. In the
        # case of distributed the vpn service should reside
        # only on a dvr_snat node.
        agent_mode = agent_conf.get(
            'agent_mode', lib_constants.L3_AGENT_MODE_LEGACY)
        if (not agent.admin_state_up or
                agent_mode == lib_constants.L3_AGENT_MODE_DVR):
            return []
        query = context.session.query(vpn_models.VPNService)
        query = query.join(vpn_models.IPsecSiteConnection)
        query = query.join(l3agent.RouterL3AgentBinding,
                           l3agent.RouterL3AgentBinding.router_id ==
                           vpn_models.VPNService.router_id)
        query = query.join(
            servicetype.ProviderResourceAssociation,
            servicetype.ProviderResourceAssociation.resource_id ==
            vpn_models.VPNService.id)
        query = query.filter(
            l3agent.RouterL3AgentBinding.l3_agent_id == agent.id)
        query = query.filter(
            servicetype.ProviderResourceAssociation.provider_name ==
            self.driver.name)
        return query

    @db_api.CONTEXT_WRITER
    def get_vpn_services_on_host(self, context, host=None):
        """Returns the vpnservices on the host."""
        vpnservices = self._get_agent_hosting_vpn_services(
            context, host)
        plugin = self.driver.service_plugin
        local_cidr_map = plugin._build_local_subnet_cidr_map(context)
        return [self.driver.make_vpnservice_dict(vpnservice, local_cidr_map)
                for vpnservice in vpnservices]

    def update_status(self, context, status):
        """Update status of vpnservices."""
        plugin = self.driver.service_plugin
        plugin.update_status_by_agent(context, status)


class IPsecVpnAgentApi(service_drivers.BaseIPsecVpnAgentApi):
    """Agent RPC API for IPsecVPNAgent."""

    target = oslo_messaging.Target(version=BASE_IPSEC_VERSION)

    # pylint: disable=useless-super-delegation
    def __init__(self, topic, default_version, driver):
        super().__init__(topic, default_version, driver)


class BaseIPsecVPNDriver(service_drivers.VpnDriver, metaclass=abc.ABCMeta):
    """Base VPN Service Driver class."""

    def __init__(self, service_plugin, validator=None):
        super().__init__(service_plugin, validator)
        self.create_rpc_conn()

    @property
    def service_type(self):
        return IPSEC

    @abc.abstractmethod
    def create_rpc_conn(self):
        pass

    def create_ipsec_site_connection(self, context, ipsec_site_connection):
        router_id = self.service_plugin.get_vpnservice_router_id(
            context, ipsec_site_connection['vpnservice_id'])
        self.agent_rpc.vpnservice_updated(context, router_id)

    def update_ipsec_site_connection(
        self, context, old_ipsec_site_connection, ipsec_site_connection):
        router_id = self.service_plugin.get_vpnservice_router_id(
            context, ipsec_site_connection['vpnservice_id'])
        self.agent_rpc.vpnservice_updated(context, router_id)

    def delete_ipsec_site_connection(self, context, ipsec_site_connection):
        router_id = self.service_plugin.get_vpnservice_router_id(
            context, ipsec_site_connection['vpnservice_id'])
        self.agent_rpc.vpnservice_updated(context, router_id)

    def create_ikepolicy(self, context, ikepolicy):
        pass

    def delete_ikepolicy(self, context, ikepolicy):
        pass

    def update_ikepolicy(self, context, old_ikepolicy, ikepolicy):
        pass

    def create_ipsecpolicy(self, context, ipsecpolicy):
        pass

    def delete_ipsecpolicy(self, context, ipsecpolicy):
        pass

    def update_ipsecpolicy(self, context, old_ipsec_policy, ipsecpolicy):
        pass

    def _get_gateway_ips(self, router):
        """Obtain the IPv4 and/or IPv6 GW IP for the router.

        If there are multiples, (arbitrarily) use the first one.
        """
        v4_ip = v6_ip = None
        for fixed_ip in router.gw_port['fixed_ips']:
            addr = fixed_ip['ip_address']
            vers = netaddr.IPAddress(addr).version
            if vers == 4:
                if v4_ip is None:
                    v4_ip = addr
            elif v6_ip is None:
                v6_ip = addr
        return v4_ip, v6_ip

    @db_api.CONTEXT_WRITER
    def create_vpnservice(self, context, vpnservice_dict):
        """Get the gateway IP(s) and save for later use.

        For the reference implementation, this side's tunnel IP (external_ip)
        will be the router's GW IP. IPSec connections will use a GW IP of
        the same version, as is used for the peer, so we will collect the
        first IP for each version (if they exist) and save them.
        """
        vpnservice = self.service_plugin._get_vpnservice(context,
                                                         vpnservice_dict['id'])
        v4_ip, v6_ip = self._get_gateway_ips(vpnservice.router)
        vpnservice_dict['external_v4_ip'] = v4_ip
        vpnservice_dict['external_v6_ip'] = v6_ip
        self.service_plugin.set_external_tunnel_ips(context,
                                                    vpnservice_dict['id'],
                                                    v4_ip=v4_ip, v6_ip=v6_ip)

    def update_vpnservice(self, context, old_vpnservice, vpnservice):
        self.agent_rpc.vpnservice_updated(context, vpnservice['router_id'])

    def delete_vpnservice(self, context, vpnservice):
        self.agent_rpc.vpnservice_updated(context, vpnservice['router_id'])

    def get_external_ip_based_on_peer(self, vpnservice, ipsec_site_con):
        """Use service's external IP, based on peer IP version."""
        vers = netaddr.IPAddress(ipsec_site_con['peer_address']).version
        if vers == 4:
            ip_to_use = vpnservice.external_v4_ip
        else:
            ip_to_use = vpnservice.external_v6_ip
        # TODO(pcm): Add validator to check that connection's peer address has
        # a version that is available in service table, so can fail early and
        # don't need a check here.
        return ip_to_use

    def make_vpnservice_dict(self, vpnservice, local_cidr_map):
        """Convert vpnservice information for vpn agent.

        also converting parameter name for vpn agent driver
        """
        vpnservice_dict = dict(vpnservice)
        # Populate tenant_id for RPC compat
        vpnservice_dict['tenant_id'] = vpnservice_dict['project_id']
        vpnservice_dict['ipsec_site_connections'] = []
        if vpnservice.subnet:
            vpnservice_dict['subnet'] = dict(vpnservice.subnet)
        else:
            vpnservice_dict['subnet'] = None
            # NOTE: Following is used for rolling upgrades, where agent may be
            # at version N, and server at N+1. We need to populate the subnet
            # with (only) the CIDR from the first connection's local endpoint
            # group.
            subnet_cidr = None
        # Not removing external_ip from vpnservice_dict, as some providers
        # may be still using it from vpnservice_dict. Will use whichever IP
        # is specified.
        vpnservice_dict['external_ip'] = (
            vpnservice.external_v4_ip or vpnservice.external_v6_ip)
        for ipsec_site_connection in vpnservice.ipsec_site_connections:
            ipsec_site_connection_dict = dict(ipsec_site_connection)
            try:
                netaddr.IPAddress(ipsec_site_connection_dict['peer_id'])
                if ipsec_site_connection_dict['local_id']:
                    netaddr.IPAddress(ipsec_site_connection_dict['local_id'])
            except netaddr.core.AddrFormatError:
                ipsec_site_connection_dict['peer_id'] = (
                    '@' + ipsec_site_connection_dict['peer_id'])
                if ipsec_site_connection_dict['local_id']:
                    ipsec_site_connection_dict['local_id'] = (
                        '@' + ipsec_site_connection_dict['local_id'])
            ipsec_site_connection_dict['ikepolicy'] = dict(
                ipsec_site_connection.ikepolicy)
            ipsec_site_connection_dict['ipsecpolicy'] = dict(
                ipsec_site_connection.ipsecpolicy)
            vpnservice_dict['ipsec_site_connections'].append(
                ipsec_site_connection_dict)
            if vpnservice.subnet:
                local_cidrs = [vpnservice.subnet.cidr]
                peer_cidrs = [
                    peer_cidr.cidr
                    for peer_cidr in ipsec_site_connection.peer_cidrs]
            else:
                local_cidrs = [
                    local_cidr_map[ep.endpoint]
                    for ep in ipsec_site_connection.local_ep_group.endpoints]
                peer_cidrs = [
                    ep.endpoint
                    for ep in ipsec_site_connection.peer_ep_group.endpoints]
                if not subnet_cidr:
                    epg = ipsec_site_connection.local_ep_group
                    subnet_cidr = local_cidr_map[epg.endpoints[0].endpoint]
            ipsec_site_connection_dict['peer_cidrs'] = peer_cidrs
            ipsec_site_connection_dict['local_cidrs'] = local_cidrs
            ipsec_site_connection_dict['local_ip_vers'] = netaddr.IPNetwork(
                local_cidrs[0]).version
            ipsec_site_connection_dict['external_ip'] = (
                self.get_external_ip_based_on_peer(vpnservice,
                                                   ipsec_site_connection_dict))
        if not vpnservice.subnet:
            vpnservice_dict['subnet'] = {'cidr': subnet_cidr}

        return vpnservice_dict



# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron.tests import base
from neutron_vpnaas.services.vpn.common import netns_wrapper as nswrap


class TestNetnsWrapper(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        patch_methods = ['filter_command',
                         'execute',
                         'setup_conf']
        for method in patch_methods:
            self.patch_obj(nswrap, method)
        patch_classes = ['neutron.common.config.setup_logging',
                         'os.path.isdir',
                         'os.path.samefile',
                         'sys.exit']
        for cls in patch_classes:
            self.patch_cls(cls)

        self.filter_command.return_value = False
        self.execute.return_value = 0
        self.conf = mock.Mock()
        self.conf.cmd = 'ls,-al'
        self.conf.mount_paths = {'/foo': '/dir/foo',
                                 '/var': '/dir/var'}
        self.setup_conf.return_value = self.conf
        self.conf.rootwrap_config = 'conf'
        self.isdir.return_value = True
        self.samefile.return_value = False

    def patch_obj(self, obj, method):
        _m = mock.patch.object(obj, method)
        _mock = _m.start()
        setattr(self, method, _mock)

    def patch_cls(self, patch_class):
        _m = mock.patch(patch_class)
        mock_name = patch_class.split('.')[-1]
        _mock = _m.start()
        setattr(self, mock_name, _mock)

    def test_netns_wrap_fail_without_netns(self):
        self.samefile.return_value = True
        return_val = nswrap.execute_with_mount()
        self.assertTrue(return_val)

    def test_netns_wrap(self):
        self.conf.cmd = 'ls,-al'
        return_val = nswrap.execute_with_mount()
        exp_calls = [mock.call(['mount', '--bind', '/dir/foo', '/foo']),
                     mock.call(['mount', '--bind', '/dir/var', '/var']),
                     mock.call('ls,-al')]
        self.execute.assert_has_calls(exp_calls, any_order=True)
        self.assertFalse(return_val)

    def test_netns_wrap_fail_without_cmd(self):
        self.conf.cmd = None
        return_val = nswrap.execute_with_mount()
        self.assertFalse(self.execute.called)
        self.assertTrue(return_val)

    def test_netns_wrap_fail_without_mount_paths(self):
        self.conf.mount_paths = None
        return_val = nswrap.execute_with_mount()
        self.assertFalse(self.execute.called)
        self.assertTrue(return_val)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import concurrent.futures
import re
import threading
import time

from oslo_utils import uuidutils
from rally.common import logging
from rally.plugins.openstack import scenario as rally_base
from rally.task import atomic

import vpn_utils

LOG = logging.getLogger(__name__)
LOCK = threading.RLock()
MAX_RESOURCES = 2


class VpnBase(rally_base.OpenStackScenario):

    def setup(self, **kwargs):
        """Create and initialize data structures to hold various resources"""

        with LOCK:
            LOG.debug('SETUP RESOURCES')
            self.neutron_admin_client = self.admin_clients("neutron")
            if kwargs['use_admin_client']:
                self.neutron_client = self.neutron_admin_client
                self.keystone_client = self.admin_clients("keystone")
                self.nova_client = self.admin_clients("nova")
            else:
                self.neutron_client = self.clients("neutron")
                self.nova_client = self.clients("nova")
            self.suffixes = [uuidutils.generate_uuid(),
                             uuidutils.generate_uuid()]
            self.remote_key_files = ['rally_keypair_' + x
                                     for x in self.suffixes]
            self.local_key_files = ['/tmp/' + x for x in self.remote_key_files]
            self.private_key_file = kwargs["private_key"]
            self.keypairs = []
            self.tenant_ids = []
            self.ns_controller_tuples = []
            self.qrouterns_compute_tuples = []
            self.router_ids = []
            self.rally_router_gw_ips = []
            self.rally_routers = []
            self.rally_networks = []
            self.rally_subnets = []
            self.rally_cidrs = []
            self.ike_policy = None
            self.ipsec_policy = None
            self.vpn_services = []
            self.ipsec_site_connections = []
            self.servers = []
            self.server_private_ips = []
            self.server_fips = []

    def create_tenants(self):
        """Create tenants"""

        for x in range(MAX_RESOURCES):
            tenant_id = vpn_utils.create_tenant(
                self.keystone_client, self.suffixes[x])
            with LOCK:
                self.tenant_ids.append(tenant_id)

    def create_networks(self, **kwargs):
        """Create networks to test vpn connectivity"""

        for x in range(MAX_RESOURCES):
            if self.tenant_ids:
                router, network, subnet, cidr = vpn_utils.create_network(
                    self.neutron_client, self.neutron_admin_client,
                    self.suffixes[x], tenant_id=self.tenant_ids[x],
                    DVR_flag=kwargs["DVR_flag"],
                    ext_net_name=kwargs["ext-net"])
            else:
                router, network, subnet, cidr = vpn_utils.create_network(
                    self.neutron_client, self.neutron_admin_client,
                    self.suffixes[x], DVR_flag=kwargs["DVR_flag"],
                    ext_net_name=kwargs["ext-net"])
            with LOCK:
                self.rally_cidrs.append(cidr)
                self.rally_subnets.append(subnet)
                self.rally_networks.append(network)
                self.rally_routers.append(router)
                self.router_ids.append(router["router"]['id'])
                self.rally_router_gw_ips.append(
                    router["router"]["external_gateway_info"]
                    ["external_fixed_ips"][0]["ip_address"])

            if kwargs["DVR_flag"]:
                ns, controller = vpn_utils.wait_for_namespace_creation(
                    "snat-", router["router"]['id'],
                    kwargs['controller_creds'],
                    self.private_key_file,
                    kwargs['namespace_creation_timeout'])
            else:
                ns, controller = vpn_utils.wait_for_namespace_creation(
                    "qrouter-", router["router"]['id'],
                    kwargs['controller_creds'],
                    self.private_key_file,
                    kwargs['namespace_creation_timeout'])
            with LOCK:
                self.ns_controller_tuples.append((ns, controller))

    def create_servers(self, **kwargs):
        """Create servers"""

        for x in range(MAX_RESOURCES):
            kwargs.update({
                "nics":
                    [{"net-id": self.rally_networks[x]["network"]["id"]}],
                "sec_group_suffix": self.suffixes[x],
                "server_suffix": self.suffixes[x]
            })
            keypair = vpn_utils.create_keypair(
                self.nova_client, self.suffixes[x])
            server = vpn_utils.create_server(
                self.nova_client, keypair, **kwargs)
            vpn_utils.assert_server_status(server, **kwargs)
            with LOCK:
                self.servers.append(server)
                self.keypairs.append(keypair)
                self.server_private_ips.append(vpn_utils.get_server_ip(
                    self.nova_client, server.id, self.suffixes[x]))
            if kwargs["DVR_flag"]:
                qrouter, compute = vpn_utils.wait_for_namespace_creation(
                    "qrouter-", self.router_ids[x],
                    kwargs['compute_creds'],
                    self.private_key_file,
                    kwargs['namespace_creation_timeout'])

                vpn_utils.write_key_to_compute_node(
                    keypair, self.local_key_files[x],
                    self.remote_key_files[x], compute,
                    self.private_key_file)
                with LOCK:
                    self.qrouterns_compute_tuples.append((qrouter, compute))
            else:
                vpn_utils.write_key_to_local_path(self.keypairs[x],
                                                  self.local_key_files[x])
                fip = vpn_utils.add_floating_ip(self.nova_client, server)
                with LOCK:
                    self.server_fips.append(fip)

    def check_route(self):
        """Verify route exists between the router gateways"""

        LOG.debug("VERIFY ROUTE EXISTS BETWEEN THE ROUTER GATEWAYS")
        for tuple in self.ns_controller_tuples:
            for ip in self.rally_router_gw_ips:
                assert vpn_utils.ping_router_gateway(
                        tuple, ip, self.private_key_file), (
                        "PING TO IP " + ip + " FAILED")

    @atomic.action_timer("_create_ike_policy")
    def _create_ike_policy(self, **kwargs):
        """Create IKE policy

        :return: IKE policy
        """
        LOG.debug('CREATING IKE_POLICY')
        ike_policy = self.neutron_client.create_ikepolicy({
            "ikepolicy": {
                "phase1_negotiation_mode":
                    kwargs.get("phase1_negotiation_mode", "main"),
                "auth_algorithm": kwargs.get("auth_algorithm", "sha1"),
                "encryption_algorithm":
                    kwargs.get("encryption_algorithm", "aes-128"),
                "pfs": kwargs.get("pfs", "group5"),
                "lifetime": {
                    "units": "seconds",
                    "value": kwargs.get("value", 7200)},
                "ike_version": kwargs.get("ike_version", "v1"),
                "name": "rally_ikepolicy"
            }
        })
        return ike_policy

    @atomic.action_timer("_create_ipsec_policy")
    def _create_ipsec_policy(self, **kwargs):
        """Create IPSEC policy

        :return: IPSEC policy
        """
        LOG.debug('CREATING IPSEC_POLICY')
        ipsec_policy = self.neutron_client.create_ipsecpolicy({
            "ipsecpolicy": {
                "name": "rally_ipsecpolicy",
                "transform_protocol": kwargs.get("transform_protocol", "esp"),
                "auth_algorithm": kwargs.get("auth_algorithm", "sha1"),
                "encapsulation_mode":
                    kwargs.get("encapsulation_mode", "tunnel"),
                "encryption_algorithm":
                    kwargs.get("encryption_algorithm", "aes-128"),
                "pfs": kwargs.get("pfs", "group5"),
                "lifetime": {
                    "units": "seconds",
                    "value": kwargs.get("value", 7200)
                }
            }
        })
        return ipsec_policy

    @atomic.action_timer("_create_vpn_service")
    def _create_vpn_service(self, rally_subnet, rally_router, vpn_suffix=None):
        """Create VPN service endpoints

        :param rally_subnet: local subnet
        :param rally_router: router endpoint
        :param vpn_suffix: suffix name for vpn service
        :return: VPN service
        """
        LOG.debug('CREATING VPN_SERVICE')
        vpn_service = self.neutron_client.create_vpnservice({
            "vpnservice": {
                "subnet_id": rally_subnet["subnet"]["id"],
                "router_id": rally_router["router"]["id"],
                "name": "rally_vpn_service_" + vpn_suffix,
                "admin_state_up": True
            }
        })
        return vpn_service

    def create_vpn_services(self):
        """Create VPN services"""

        for x in range(MAX_RESOURCES):
            vpn_service = self._create_vpn_service(
                self.rally_subnets[x], self.rally_routers[x], self.suffixes[x])
            with LOCK:
                self.vpn_services.append(vpn_service)

    @atomic.action_timer("_create_ipsec_site_connection")
    def _create_ipsec_site_connection(self, local_index, peer_index, **kwargs):
        """Create IPSEC site connection

        :param local_index: parameter to point to the local end-point
        :param peer_index: parameter to point to the peer end-point
        :return: IPSEC site connection
        """
        LOG.debug('CREATING IPSEC_SITE_CONNECTION')
        ipsec_site_conn = self.neutron_client.create_ipsec_site_connection({
            "ipsec_site_connection": {
                "psk": kwargs.get("secret", "secret"),
                "initiator": "bi-directional",
                "ipsecpolicy_id": self.ipsec_policy["ipsecpolicy"]["id"],
                "admin_state_up": True,
                "peer_cidrs": self.rally_cidrs[peer_index],
                "mtu": kwargs.get("mtu", "1500"),
                "ikepolicy_id": self.ike_policy["ikepolicy"]["id"],
                "dpd": {
                    "action": "disabled",
                    "interval": 60,
                    "timeout": 240
                },
                "vpnservice_id":
                    self.vpn_services[local_index]["vpnservice"]["id"],
                "peer_address": self.rally_router_gw_ips[peer_index],
                "peer_id": self.rally_router_gw_ips[peer_index],
                "name": "rally_ipsec_site_connection_" +
                        self.suffixes[local_index]
            }
        })
        return ipsec_site_conn

    def create_ipsec_site_connections(self, **kwargs):
        """Create IPSEC site connections"""

        a = self._create_ipsec_site_connection(0, 1, **kwargs)
        b = self._create_ipsec_site_connection(1, 0, **kwargs)
        with LOCK:
            self.ipsec_site_connections = [a, b]

    def _get_resource(self, resource_tag, resource_id):
        """Get the resource(vpn_service or ipsec_site_connection)

        :param resource_tag: "vpnservice" or "ipsec_site_connection"
        :param resource_id: id of the resource
        :return: resource (vpn_service or ipsec_site_connection)
        """
        if resource_tag == "vpnservice":
            vpn_service = self.neutron_client.show_vpnservice(resource_id)
            if vpn_service:
                return vpn_service
        elif resource_tag == 'ipsec_site_connection':
            ipsec_site_conn = self.neutron_client.show_ipsec_site_connection(
                resource_id)
            if ipsec_site_conn:
                return ipsec_site_conn

    def _wait_for_status_change(self, resource, resource_tag, final_status,
                                wait_timeout=60, check_interval=1):
        """Wait for resource's status change

        Wait till the status of the resource changes to final state or till
        the time exceeds the wait_timeout value.
        :param resource: resource whose status has to be checked
        :param final_status: desired final status of the resource
        :param resource_tag: to identify the resource as vpnservice or
                             ipsec_site_connection
        :param wait_timeout: timeout value in seconds
        :param check_interval: time to sleep before each check for the status
        change
        :return: resource
        """
        LOG.debug('WAIT_FOR_%s_STATUS_CHANGE ', resource[resource_tag]['id'])

        start_time = time.time()
        while True:
            resource = self._get_resource(
                resource_tag, resource[resource_tag]['id'])
            current_status = resource[resource_tag]['status']
            if current_status == final_status:
                return resource
            time.sleep(check_interval)
            if time.time() - start_time > wait_timeout:
                raise Exception(
                    "Timeout waiting for resource {} to change to {} status".
                    format(resource[resource_tag]['name'], final_status))

    @atomic.action_timer("wait_time_for_status_change")
    def _assert_statuses(self, ipsec_site_conn, vpn_service,
                         final_status, **kwargs):
        """Assert statuses of vpn_service and ipsec_site_connection

        :param ipsec_site_conn: ipsec_site_connection object
        :param vpn_service: vpn_service object
        :param final_status: status of vpn and ipsec_site_connection object
        """

        vpn_service = self._wait_for_status_change(
            vpn_service,
            resource_tag="vpnservice",
            final_status=final_status,
            wait_timeout=kwargs.get("vpn_service_creation_timeout"),
            check_interval=5)

        ipsec_site_conn = self._wait_for_status_change(
            ipsec_site_conn,
            resource_tag="ipsec_site_connection",
            final_status=final_status,
            wait_timeout=kwargs.get("ipsec_site_connection_creation_timeout"),
            check_interval=5)

        LOG.debug("VPN SERVICE STATUS %s", vpn_service['vpnservice']['status'])
        LOG.debug("IPSEC_SITE_CONNECTION STATUS %s",
                  ipsec_site_conn['ipsec_site_connection']['status'])

    def assert_statuses(self, final_status, **kwargs):
        """Assert active statuses for VPN services and VPN connections

        :param final_status: the final status you expect the resource to be in
        """

        LOG.debug("ASSERTING ACTIVE STATUSES FOR VPN-SERVICES AND "
                  "IPSEC-SITE-CONNECTIONS")
        for x in range(MAX_RESOURCES):
            self._assert_statuses(
                self.ipsec_site_connections[x], self.vpn_services[x],
                final_status, **kwargs)

    def _get_qg_interface(self, peer_index):
        """Get the qg- interface

        :param peer_index: parameter to point to the local end-point
        :return: qg-interface
        """
        qg = vpn_utils.get_interfaces(
            self.ns_controller_tuples[peer_index],
            self.private_key_file)
        p = re.compile(r"qg-\w+-\w+")
        for line in qg:
            m = p.search(line)
            if m:
                return m.group()
        return None

    @atomic.action_timer("_verify_vpn_connection")
    def _verify_vpn_connectivity(self, local_index, peer_index, **kwargs):
        """Verify the vpn connectivity between the endpoints

        Get the qg- interface from the snat namespace corresponding to the
        peer router and start a tcp dump. Concurrently, SSH into the nova
        instance  on the local subnet from the qrouter namespace and try
        to ping the nova instance on the peer subnet. Inspect the captured
        packets to see if they are encrypted.
        :param local_index: parameter to point to the local end-point
        :param peer_index: parameter to point to the peer end-point
        :return: True if vpn connectivity test passes
                 False if the test fails
        """
        qg_interface = self._get_qg_interface(peer_index)
        if qg_interface:
            with concurrent.futures.ThreadPoolExecutor(max_workers=2) as e:
                tcpdump_future = e.submit(vpn_utils.start_tcpdump,
                         self.ns_controller_tuples[peer_index],
                         qg_interface, self.private_key_file)
                if kwargs["DVR_flag"]:
                    ssh_future = e.submit(
                        vpn_utils.ssh_and_ping_server,
                        self.server_private_ips[local_index],
                        self.server_private_ips[peer_index],
                        self.qrouterns_compute_tuples[local_index],
                        self.remote_key_files[local_index],
                        self.private_key_file)
                else:
                    ssh_future = e.submit(
                        vpn_utils.ssh_and_ping_server_with_fip,
                        self.server_fips[local_index],
                        self.server_private_ips[peer_index],
                        self.local_key_files[local_index],
                        self.private_key_file)

            assert ssh_future.result(), "SSH/Ping failed"
            for line in tcpdump_future.result():
                if 'ESP' in line:
                    return True
        return False

    def verify_vpn_connectivity(self, **kwargs):
        """Verify VPN connectivity"""

        LOG.debug("VERIFY THE VPN CONNECTIVITY")
        with LOCK:
            assert self._verify_vpn_connectivity(
                0, 1, **kwargs), "VPN CONNECTION FAILED"
        with LOCK:
            assert self._verify_vpn_connectivity(
                1, 0, **kwargs), "VPN CONNECTION FAILED"

    def update_router(self, router_id, admin_state_up=False):
        """Update router's admin_state_up field

        :param router_id: uuid of the router
        :param admin_state_up: True or False
        """
        LOG.debug('UPDATE ROUTER')
        router_args = {'router': {'admin_state_up': admin_state_up}}
        self.neutron_client.update_router(router_id, router_args)

    @atomic.action_timer("_delete_ipsec_site_connection")
    def _delete_ipsec_site_connections(self):
        """Delete IPSEC site connections"""

        for site_conn in self.ipsec_site_connections:
            LOG.debug("DELETING IPSEC_SITE_CONNECTION %s",
                      site_conn['ipsec_site_connection']['id'])
            self.neutron_client.delete_ipsec_site_connection(
                site_conn['ipsec_site_connection']['id'])

    @atomic.action_timer("_delete_vpn_service")
    def _delete_vpn_services(self):
        """Delete VPN service endpoints"""

        for vpn_service in self.vpn_services:
            LOG.debug("DELETING VPN_SERVICE %s",
                      vpn_service['vpnservice']['id'])
            self.neutron_client.delete_vpnservice(
                vpn_service['vpnservice']['id'])

    @atomic.action_timer("_delete_ipsec_policy")
    def _delete_ipsec_policy(self):
        """Delete IPSEC policy"""

        LOG.debug("DELETING IPSEC POLICY")
        if self.ipsec_policy:
            self.neutron_client.delete_ipsecpolicy(
                self.ipsec_policy['ipsecpolicy']['id'])

    @atomic.action_timer("_delete_ike_policy")
    def _delete_ike_policy(self):
        """Delete IKE policy"""

        LOG.debug('DELETING IKE POLICY')
        if self.ike_policy:
            self.neutron_client.delete_ikepolicy(
                self.ike_policy['ikepolicy']['id'])

    @atomic.action_timer("cleanup")
    def cleanup(self):
        """Clean the resources"""

        vpn_utils.delete_servers(self.nova_client, self.servers)
        if self.server_fips:
            vpn_utils.delete_floating_ips(self.nova_client, self.server_fips)
        vpn_utils.delete_keypairs(self.nova_client, self.keypairs)

        if self.qrouterns_compute_tuples:
            vpn_utils.delete_hosts_from_knownhosts_file(
                self.server_private_ips, self.qrouterns_compute_tuples,
                self.private_key_file)
            vpn_utils.delete_keyfiles(
                self.local_key_files, self.remote_key_files,
                self.qrouterns_compute_tuples, self.private_key_file)
        else:
            vpn_utils.delete_hosts_from_knownhosts_file(
                    self.server_private_ips)
            vpn_utils.delete_keyfiles(self.local_key_files)

        self._delete_ipsec_site_connections()
        self._delete_vpn_services()
        self._delete_ipsec_policy()
        self._delete_ike_policy()
        vpn_utils.delete_networks(
            self.neutron_client, self.neutron_admin_client, self.rally_routers,
            self.rally_networks, self.rally_subnets)
        if self.tenant_ids:
            vpn_utils.delete_tenants(self.keystone_client, self.tenant_ids)



# Copyright 2016 MingShuang Xian/IBM
# Copyright 2023 SysEleven GmbH
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Add table for vpn gateway (gateway port and transit network)

Revision ID: 22e0145ac80b
Revises: 3b739d6906cf
Create Date: 2016-09-18 09:01:18.660362

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '22e0145ac80b'
down_revision = '3b739d6906cf'


def upgrade():
    op.create_table(
        'vpn_ext_gws',
        sa.Column('id', sa.String(length=36), nullable=False,
                  primary_key=True),
        sa.Column('project_id', sa.String(length=255),
                  index=True),
        sa.Column('router_id', sa.String(length=36), nullable=False,
                  unique=True),
        sa.Column('status', sa.String(length=16), nullable=False),
        sa.Column('gw_port_id', sa.String(length=36)),
        sa.Column('transit_port_id', sa.String(length=36)),
        sa.Column('transit_network_id', sa.String(length=36)),
        sa.Column('transit_subnet_id', sa.String(length=36)),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['router_id'], ['routers.id']),
        sa.ForeignKeyConstraint(['gw_port_id'], ['ports.id'],
                                ondelete='SET NULL'),
        sa.ForeignKeyConstraint(['transit_port_id'], ['ports.id'],
                                ondelete='SET NULL'),
        sa.ForeignKeyConstraint(['transit_network_id'], ['networks.id'],
                                ondelete='SET NULL'),
        sa.ForeignKeyConstraint(['transit_subnet_id'], ['subnets.id'],
                                ondelete='SET NULL'),
    )



# Copyright 2013, Nachi Ueno, NTT I3, Inc.
# Copyright 2017, Fujitsu Limited
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from neutron_lib.agent import l3_extension
from oslo_config import cfg
from oslo_log import log as logging

from neutron_vpnaas._i18n import _
from neutron_vpnaas.services.vpn import vpn_service

LOG = logging.getLogger(__name__)

vpn_agent_opts = [
    cfg.MultiStrOpt(
        'vpn_device_driver',
        default=['neutron_vpnaas.services.vpn.device_drivers.'
                 'ipsec.OpenSwanDriver'],
        sample_default=['neutron_vpnaas.services.vpn.device_drivers.ipsec.'
                        'OpenSwanDriver, '
                        'neutron_vpnaas.services.vpn.device_drivers.'
                        'strongswan_ipsec.StrongSwanDriver, '
                        'neutron_vpnaas.services.vpn.device_drivers.'
                        'libreswan_ipsec.LibreSwanDriver'],
        help=_("The vpn device drivers Neutron will use")),
]
cfg.CONF.register_opts(vpn_agent_opts, 'vpnagent')


class VPNAgent(l3_extension.L3AgentExtension):
    """VPNaaS Agent support to be used by Neutron L3 agent."""

    def initialize(self, connection, driver_type):
        LOG.debug("Loading VPNaaS")

    def consume_api(self, agent_api):
        LOG.debug("Loading consume_api for VPNaaS")
        self.agent_api = agent_api

    def __init__(self, host, conf):
        LOG.debug("Initializing VPNaaS agent")
        self.agent_api = None
        self.conf = conf
        self.host = host
        self.service = vpn_service.VPNService(self)
        self.device_drivers = self.service.load_device_drivers(self.host)

    def add_router(self, context, data):
        """Handles router add event"""
        ri = self.agent_api.get_router_info(data['id'])
        if ri is not None:
            for device_driver in self.device_drivers:
                device_driver.create_router(ri)
                device_driver.sync(context, [ri])
        else:
            LOG.debug("Router %s was concurrently deleted while "
                      "creating VPN for it", data['id'])

    def update_router(self, context, data):
        """Handles router update event"""
        ri = self.agent_api.get_router_info(data['id'])
        if ri is not None:
            for device_driver in self.device_drivers:
                device_driver.sync(context, [ri])
        else:
            LOG.debug("Router %s was concurrently deleted while "
                      "updating VPN for it", data['id'])

    def delete_router(self, context, data):
        """Handles router delete event"""
        for device_driver in self.device_drivers:
            device_driver.destroy_router(data['id'])

    def ha_state_change(self, context, data):
        """Enable the vpn process when router transitioned to master.

        And disable vpn process for backup router.
        """
        router_id = data['router_id']
        state = data['state']
        for device_driver in self.device_drivers:
            if router_id in device_driver.processes:
                # NOTE(mnaser): We need to update the router object so it has
                #               the new HA state so we can do status updates.
                device_driver.routers[router_id].ha_state = state

                process = device_driver.processes[router_id]
                if state in ('master', 'primary'):
                    process.enable()
                else:
                    process.disable()

    def update_network(self, context, data):
        pass


class L3WithVPNaaS(VPNAgent):

    def __init__(self, conf=None):
        if conf:
            self.conf = conf
        else:
            self.conf = cfg.CONF
        super().__init__(host=self.conf.host, conf=self.conf)



# Copyright 2013, Nachi Ueno, NTT I3, Inc.
# Copyright 2023, SysEleven GmbH
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import sys

from neutron.common import config as common_config
from neutron.conf.agent import common as agent_config
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import service

from neutron_vpnaas._i18n import _
from neutron_vpnaas.agent.ovn.vpn import agent

LOG = logging.getLogger(__name__)


VPN_AGENT_OPTS = [
    cfg.MultiStrOpt(
        'vpn_device_driver',
        default=['neutron_vpnaas.services.vpn.device_drivers.'
                 'ovn_ipsec.OvnStrongSwanDriver'],
        sample_default=['neutron_vpnaas.services.vpn.device_drivers.'
                        'ovn_ipsec.OvnStrongSwanDriver'],
        help=_("The OVN VPN device drivers Neutron will use")),
]

OVS_OPTS = [
    cfg.StrOpt('ovsdb_connection',
               default='unix:/usr/local/var/run/openvswitch/db.sock',
               regex=r'^(tcp|ssl|unix):.+',
               help=_('The connection string for the native OVSDB backend.\n'
                      'Use tcp:IP:PORT for TCP connection.\n'
                      'Use unix:FILE for unix domain socket connection.')),
    cfg.IntOpt('ovsdb_connection_timeout',
               default=180,
               help=_('Timeout in seconds for the OVSDB '
                      'connection transaction'))
]


def register_opts(conf):
    common_config.register_common_config_options()
    agent_config.register_interface_driver_opts_helper(conf)
    agent_config.register_interface_opts(conf)
    agent_config.register_availability_zone_opts_helper(conf)
    ovn_conf.register_opts()
    conf.register_opts(VPN_AGENT_OPTS, 'vpnagent')
    conf.register_opts(OVS_OPTS, 'ovs')


def main():
    register_opts(cfg.CONF)
    common_config.init(sys.argv[1:])
    agent_config.setup_logging()
    agent_config.setup_privsep()

    agt = agent.OvnVpnAgent(cfg.CONF)
    service.launch(cfg.CONF, agt, restart_method='mutate').wait()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os


def load_tests(loader, tests, pattern):
    this_dir = os.path.dirname(__file__)
    openswan_tests = loader.discover(start_dir=this_dir, pattern=pattern)
    tests.addTests(openswan_tests)

    common_dir = os.path.abspath(os.path.join(this_dir, "../common"))
    common_tests = loader.discover(start_dir=common_dir, pattern=pattern)
    tests.addTests(common_tests)
    return tests



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from rally.common import logging
from rally.task import scenario
from rally.task import types

import vpn_base

LOG = logging.getLogger(__name__)


class TestVpnTenantScenario(vpn_base.VpnBase):
    """Rally scenarios for VPNaaS"""

    @types.convert(image={"type": "glance_image"},
                   flavor={"type": "nova_flavor"})
    @scenario.configure()
    def multitenants_vpn_test(self, **kwargs):
        """Test VPN connectivity under two different tenants.

        1. Create 2 private networks with 2 different tenants, subnets, routers
        2. Create public network, subnets and GW IPs on routers, if not present
        3. Execute ip netns command and get the snat and qrouter namespaces
           (assuming we use DVR)
        4. Verify that there is a route between the router gateways by pinging
           each other from their snat namespaces
        5. Add security group rules for SSH and ICMP
        6. Start a nova instance in each of the private networks
        7. Create IKE and IPSEC policies
        8. Create VPN service at each of the routers
        9. Create IPSEC site connections at both endpoints
        10. Verify that the vpn-service and ipsec-site-connection are ACTIVE
        11. Cleanup the resources that are setup for this test
        """

        try:
            self.setup(**kwargs)
            self.create_tenants()
            self.create_networks(**kwargs)
            self.check_route()
            self.ike_policy = self._create_ike_policy(**kwargs)
            self.ipsec_policy = self._create_ipsec_policy(**kwargs)
            self.create_vpn_services()
            self.create_ipsec_site_connections(**kwargs)
            self.assert_statuses(final_status='ACTIVE', **kwargs)
            LOG.info("VPN TENANT TEST PASSED!")

        finally:
            self.cleanup()



# Copyright 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.services import provider_configuration as provconfig
from neutron_lib.exceptions import vpn as vpn_exception
from oslo_log import log as logging
from oslo_utils import importutils

LOG = logging.getLogger(__name__)

DEVICE_DRIVERS = 'device_drivers'


class VPNService:
    """VPN Service observer."""

    def __init__(self, l3_agent):
        self.conf = l3_agent.conf

    def load_device_drivers(self, host):
        """Loads one or more device drivers for VPNaaS."""
        drivers = []
        for device_driver in self.conf.vpnagent.vpn_device_driver:
            device_driver = provconfig.get_provider_driver_class(
                device_driver, DEVICE_DRIVERS)
            try:
                drivers.append(importutils.import_object(device_driver,
                                                         self,
                                                         host))
                LOG.debug('Loaded VPNaaS device driver: %s', device_driver)
            except ImportError:
                raise vpn_exception.DeviceDriverImportError(
                    device_driver=device_driver)
        return drivers



# Copyright 2016, Yi Jing Zhu, IBM.
# Copyright 2023, SysEleven GmbH
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from neutron_lib.api.definitions import portbindings
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as lib_constants
from neutron_lib import context as nctx
from neutron_lib.db import api as db_api
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from neutron_lib.plugins import utils as p_utils
from neutron_lib import rpc as n_rpc

from oslo_config import cfg
from oslo_db import exception as o_exc
from oslo_log import log as logging

from neutron_vpnaas.db.vpn import vpn_agentschedulers_db as agent_db
from neutron_vpnaas.db.vpn.vpn_ext_gw_db import RouterIsNotVPNExternal
from neutron_vpnaas.db.vpn import vpn_models
from neutron_vpnaas.extensions import vpnaas
from neutron_vpnaas.services.vpn.common import topics
from neutron_vpnaas.services.vpn.service_drivers import base_ipsec


LOG = logging.getLogger(__name__)

IPSEC = 'ipsec'
BASE_IPSEC_VERSION = '1.0'

TRANSIT_NETWORK_PREFIX = 'vpn-transit-network-'
TRANSIT_SUBNET_PREFIX = 'vpn-transit-subnet-'
TRANSIT_PORT_PREFIX = 'vpn-ns-'
VPN_GW_PORT_PREFIX = 'vpn-gw-'
VPN_TRANSIT_LIP = '169.254.0.1'
VPN_TRANSIT_RIP = '169.254.0.2'
VPN_TRANSIT_CIDR = '169.254.0.0/28'
HIDDEN_PROJECT_ID = ''


class IPsecVpnOvnDriverCallBack(base_ipsec.IPsecVpnDriverCallBack):
    def __init__(self, driver):
        super().__init__(driver)
        self.admin_ctx = nctx.get_admin_context()

    @property
    def core_plugin(self):
        return self.driver.core_plugin

    @property
    def service_plugin(self):
        return self.driver.service_plugin

    def _get_vpn_gateway(self, context, router_id):
        return self.service_plugin.get_vpn_gw_by_router_id(context, router_id)

    def get_vpn_transit_network_details(self, context, router_id):
        vpn_gw = self._get_vpn_gateway(context, router_id)
        network_id = vpn_gw.gw_port['network_id']
        external_network = self.core_plugin.get_network(context, network_id)

        details = {
            'gw_port': vpn_gw.gw_port,
            'transit_port': vpn_gw.transit_port,
            'transit_gateway_ip': VPN_TRANSIT_LIP,
            'external_network': external_network,
        }
        return details

    def get_subnet_info(self, context, subnet_id=None):
        try:
            return self.core_plugin.get_subnet(context, subnet_id)
        except n_exc.SubnetNotFound:
            return None

    def _get_agent_hosting_vpn_services(self, context, host):
        agent = self.service_plugin.get_vpn_agent_on_host(context, host)
        if not agent:
            return []

        # We're here because a VPN agent asked for the VPN services it's
        # hosting. This means, the agent is alive. This is a chance to
        # schedule VPN services of routers that are still unscheduled.
        if cfg.CONF.vpn_auto_schedule:
            self.service_plugin.auto_schedule_routers(context, agent)

        query = context.session.query(vpn_models.VPNService)
        query = query.join(vpn_models.IPsecSiteConnection)
        query = query.join(agent_db.RouterVPNAgentBinding,
                           agent_db.RouterVPNAgentBinding.router_id ==
                           vpn_models.VPNService.router_id)
        query = query.filter(
            agent_db.RouterVPNAgentBinding.vpn_agent_id == agent['id'])
        return query


@registry.has_registry_receivers
class BaseOvnIPsecVPNDriver(base_ipsec.BaseIPsecVPNDriver):
    def __init__(self, service_plugin):
        self._l3_plugin = None
        self._core_plugin = None
        super().__init__(service_plugin)

    @property
    def l3_plugin(self):
        if self._l3_plugin is None:
            self._l3_plugin = directory.get_plugin(plugin_constants.L3)
        return self._l3_plugin

    @property
    def core_plugin(self):
        if self._core_plugin is None:
            self._core_plugin = directory.get_plugin()
        return self._core_plugin

    @registry.receives(resources.ROUTER, [events.PRECOMMIT_UPDATE])
    def _handle_router_precommit_update(self, resource, event, trigger,
                                        payload):
        """Check that a router update won't remove routes we need for VPN."""
        LOG.debug("Router %s PRECOMMIT_UPDATE event: %s",
                  payload.resource_id, payload.request_body)
        router_id = payload.resource_id
        context = payload.context
        router_data = payload.request_body
        routes_removed = router_data.get('routes_removed')

        if not routes_removed:
            return

        removed_cidrs = {r['destination'] for r in routes_removed}
        vpn_cidrs = set(
            self.service_plugin.get_peer_cidrs_for_router(context, router_id))
        conflict_cidrs = removed_cidrs.intersection(vpn_cidrs)

        if conflict_cidrs:
            raise vpnaas.RouteInUseByVPN(
                destinations=", ".join(conflict_cidrs))

    def get_vpn_gw_port_name(self, router_id):
        return VPN_GW_PORT_PREFIX + router_id

    def get_vpn_namespace_port_name(self, router_id):
        return TRANSIT_PORT_PREFIX + router_id

    def get_transit_network_name(self, router_id):
        return TRANSIT_NETWORK_PREFIX + router_id

    def get_transit_subnet_name(self, router_id):
        return TRANSIT_SUBNET_PREFIX + router_id

    def make_transit_network(self, router_id, tenant_id, agent_host,
                             gateway_update):
        context = nctx.get_admin_context()
        network_data = {
            'tenant_id': HIDDEN_PROJECT_ID,
            'name': self.get_transit_network_name(router_id),
            'admin_state_up': True,
            'shared': False,
        }
        network = p_utils.create_network(self.core_plugin, context,
                                         {'network': network_data})
        gateway_update['transit_network_id'] = network['id']

        # The subnet tenant_id must be of the user, otherwise updating the
        # router by the user may fail (it needs access to all subnets)
        subnet_data = {
            'tenant_id': tenant_id,
            'name': self.get_transit_subnet_name(router_id),
            'gateway_ip': VPN_TRANSIT_LIP,
            'cidr': VPN_TRANSIT_CIDR,
            'network_id': network['id'],
            'ip_version': 4,
            'enable_dhcp': False,
        }
        subnet = p_utils.create_subnet(self.core_plugin, context,
                                       {'subnet': subnet_data})
        gateway_update['transit_subnet_id'] = subnet['id']

        self.l3_plugin.add_router_interface(context, router_id,
                                            {'subnet_id': subnet['id']})

        fixed_ip = {'subnet_id': subnet['id'], 'ip_address': VPN_TRANSIT_RIP}
        port_data = {
            'tenant_id': HIDDEN_PROJECT_ID,
            'network_id': network['id'],
            'fixed_ips': [fixed_ip],
            'device_id': subnet['id'],
            'device_owner': lib_constants.DEVICE_OWNER_VPN_TRANSIT_NETWORK,
            'admin_state_up': True,
            portbindings.HOST_ID: agent_host,
            'name': self.get_vpn_namespace_port_name(router_id)
        }
        port = p_utils.create_port(self.core_plugin, context,
                                   {"port": port_data})
        gateway_update['transit_port_id'] = port['id']

    def _del_port(self, context, port_id):
        try:
            self.core_plugin.delete_port(context, port_id, l3_port_check=False)
        except n_exc.PortNotFound:
            pass

    def _remove_router_interface(self, context, router_id, subnet_id):
        try:
            self.l3_plugin.remove_router_interface(
                context, router_id, {'subnet_id': subnet_id})
        except (n_exc.l3.RouterInterfaceNotFoundForSubnet,
                n_exc.SubnetNotFound):
            pass

    def _del_subnet(self, context, subnet_id):
        try:
            self.core_plugin.delete_subnet(context, subnet_id)
        except n_exc.SubnetNotFound:
            pass

    def _del_network(self, context, network_id):
        try:
            self.core_plugin.delete_network(context, network_id)
        except n_exc.NetworkNotFound:
            pass

    def del_transit_network(self, gw):
        context = nctx.get_admin_context()
        router_id = gw['router_id']

        port_id = gw.get('transit_port_id')
        if port_id:
            self._del_port(context, port_id)

        subnet_id = gw.get('transit_subnet_id')
        if subnet_id:
            self._remove_router_interface(context, router_id, subnet_id)
            self._del_subnet(context, subnet_id)

        network_id = gw.get('transit_network_id')
        if network_id:
            self._del_network(context, network_id)

    def make_gw_port(self, router_id, network_id, agent_host, gateway_update):
        context = nctx.get_admin_context()
        port_data = {'tenant_id': HIDDEN_PROJECT_ID,
                     'network_id': network_id,
                     'fixed_ips': lib_constants.ATTR_NOT_SPECIFIED,
                     'device_id': router_id,
                     'device_owner': lib_constants.DEVICE_OWNER_VPN_ROUTER_GW,
                     'admin_state_up': True,
                     portbindings.HOST_ID: agent_host,
                     'name': self.get_vpn_gw_port_name(router_id)}
        gw_port = p_utils.create_port(self.core_plugin, context.elevated(),
                                      {'port': port_data})

        if not gw_port['fixed_ips']:
            LOG.debug('No IPs available for external network %s', network_id)
        gateway_update['gw_port_id'] = gw_port['id']

    def del_gw_port(self, gateway):
        context = nctx.get_admin_context()
        port_id = gateway.get('gw_port_id')
        if port_id:
            self._del_port(context, port_id)

    def _get_peer_cidrs(self, vpnservice):
        cidrs = []
        for ipsec_site_connection in vpnservice.ipsec_site_connections:
            if ipsec_site_connection.peer_cidrs:
                for peer_cidr in ipsec_site_connection.peer_cidrs:
                    cidrs.append(peer_cidr.cidr)
            if ipsec_site_connection.peer_ep_group is not None:
                for ep in ipsec_site_connection.peer_ep_group.endpoints:
                    cidrs.append(ep.endpoint)
        return cidrs

    def _routes_update(self, cidrs, nexthop):
        routes = [{'destination': cidr, 'nexthop': nexthop}
                  for cidr in cidrs]
        return {'router': {'routes': routes}}

    def _update_static_routes(self, context, ipsec_site_connection):
        vpnservice = self.service_plugin.get_vpnservice(
            context, ipsec_site_connection['vpnservice_id'])
        router_id = vpnservice['router_id']
        gw = self.service_plugin.get_vpn_gw_by_router_id(context, router_id)

        nexthop = gw.transit_port['fixed_ips'][0]['ip_address']

        router = self.l3_plugin.get_router(context, router_id)
        old_routes = router.get('routes', [])

        old_cidrs = {r['destination'] for r in old_routes
                     if r['nexthop'] == nexthop}
        new_cidrs = set(
            self.service_plugin.get_peer_cidrs_for_router(context, router_id))

        to_remove = old_cidrs - new_cidrs
        if to_remove:
            self.l3_plugin.remove_extraroutes(
                context, router_id, self._routes_update(to_remove, nexthop))

        to_add = new_cidrs - old_cidrs
        if to_add:
            self.l3_plugin.add_extraroutes(
                context, router_id, self._routes_update(to_add, nexthop))

    def _get_gateway_ips(self, router):
        """Obtain the IPv4 and/or IPv6 GW IP for the router.

        If there are multiples, (arbitrarily) use the first one.
        """
        gateway = self.service_plugin.get_vpn_gw_dict_by_router_id(
            nctx.get_admin_context(),
            router['id'])
        if gateway is None or gateway['external_fixed_ips'] is None:
            raise RouterIsNotVPNExternal(router_id=router['id'])

        v4_ip = v6_ip = None
        for fixed_ip in gateway['external_fixed_ips']:
            addr = fixed_ip['ip_address']
            vers = netaddr.IPAddress(addr).version
            if vers == lib_constants.IP_VERSION_4:
                if v4_ip is None:
                    v4_ip = addr
            elif v6_ip is None:
                v6_ip = addr
        return v4_ip, v6_ip

    def _update_gateway(self, context, gateway_id, **kwargs):
        gateway = {'gateway': kwargs}
        return self.service_plugin.update_gateway(context, gateway_id, gateway)

    @db_api.retry_if_session_inactive()
    def _ensure_gateway(self, context, vpnservice):
        gw = self.service_plugin.get_vpn_gw_dict_by_router_id(
            context, vpnservice['router_id'], refresh=True)
        if not gw:
            gateway = {'gateway': {
                'router_id': vpnservice['router_id'],
                'tenant_id': vpnservice['tenant_id'],
            }}
            # create_gateway may raise oslo_db.exception.DBDuplicateEntry
            # if someone else created one in the meantime
            return self.service_plugin.create_gateway(context, gateway)

        if gw['status'] == lib_constants.ERROR:
            raise vpnaas.VPNGatewayInError()

        # Raise an exception if an existing gateway is in status
        # PENDING_CREATE or PENDING_DELETE.
        # One of the next retries should succeed.
        if gw['status'] != lib_constants.ACTIVE:
            raise o_exc.RetryRequest(vpnaas.VPNGatewayNotReady())
        return gw

    @db_api.CONTEXT_WRITER
    def _setup(self, context, vpnservice_dict):
        router_id = vpnservice_dict['router_id']
        agent = self.service_plugin.schedule_router(context, router_id)
        if not agent:
            raise vpnaas.NoVPNAgentAvailable
        agent_host = agent['host']

        gateway = self._ensure_gateway(context, vpnservice_dict)

        # If the gateway status is ACTIVE the ports have been created already
        if gateway['status'] == lib_constants.ACTIVE:
            return

        vpnservice = self.service_plugin._get_vpnservice(context,
                                                         vpnservice_dict['id'])
        network_id = vpnservice.router.gw_port.network_id
        gateway_update = {}  # keeps track of already-created IDs
        try:
            self.make_gw_port(router_id, network_id, agent_host,
                              gateway_update)
            self.make_transit_network(router_id,
                                      vpnservice_dict['tenant_id'],
                                      agent_host,
                                      gateway_update)
        except Exception:
            self._update_gateway(context, gateway['id'],
                                 status=lib_constants.ERROR,
                                 **gateway_update)
            raise

        self._update_gateway(context, gateway['id'],
                             status=lib_constants.ACTIVE,
                             **gateway_update)

    def _cleanup(self, context, router_id):
        gw = self.service_plugin.get_vpn_gw_dict_by_router_id(context,
                                                              router_id)
        if not gw:
            return
        self._update_gateway(context, gw['id'],
                             status=lib_constants.PENDING_DELETE)
        try:
            self.del_gw_port(gw)
            self.del_transit_network(gw)
            self.service_plugin.delete_gateway(context, gw['id'])
        except Exception:
            LOG.exception("Cleanup of VPN gateway for router %s failed.",
                          router_id)
            self._update_gateway(context, gw['id'],
                                 status=lib_constants.ERROR)
            raise

    def create_vpnservice(self, context, vpnservice_dict):
        try:
            self._setup(context, vpnservice_dict)
        except Exception:
            LOG.exception("Setting up the VPN gateway for router %s failed.",
                          vpnservice_dict['router_id'])
            self.service_plugin.set_vpnservice_status(
                context, vpnservice_dict['id'], lib_constants.ERROR,
                updated_pending_status=True)
            raise
        super().create_vpnservice(context, vpnservice_dict)

    def delete_vpnservice(self, context, vpnservice):
        router_id = vpnservice['router_id']
        super().delete_vpnservice(context, vpnservice)
        services = self.service_plugin.get_vpnservices(context)
        router_ids = [s['router_id'] for s in services]
        if router_id not in router_ids:
            self._cleanup(context, router_id)

    def create_ipsec_site_connection(self, context, ipsec_site_connection):
        self._update_static_routes(context, ipsec_site_connection)
        super().create_ipsec_site_connection(context, ipsec_site_connection)

    def delete_ipsec_site_connection(self, context, ipsec_site_connection):
        self._update_static_routes(context, ipsec_site_connection)
        super().delete_ipsec_site_connection(context, ipsec_site_connection)

    def update_ipsec_site_connection(
            self, context, old_ipsec_site_connection, ipsec_site_connection):
        self._update_static_routes(context, ipsec_site_connection)
        super().update_ipsec_site_connection(
            context, old_ipsec_site_connection, ipsec_site_connection)

    def _update_port_binding(self, context, port_id, host):
        port_data = {'binding:host_id': host}
        self.core_plugin.update_port(context, port_id, {'port': port_data})

    def update_port_bindings(self, context, router_id, host):
        gw = self.service_plugin.get_vpn_gw_dict_by_router_id(context,
                                                              router_id)
        if not gw:
            return
        port_id = gw.get('gw_port_id')
        if port_id:
            self._update_port_binding(context, port_id, host)
        port_id = gw.get('transit_port_id')
        if port_id:
            self._update_port_binding(context, port_id, host)


class IPsecOvnVpnAgentApi(base_ipsec.IPsecVpnAgentApi):
    def _agent_notification(self, context, method, router_id,
                            version=None, **kwargs):
        """Notify update for the agent.

        This method will find where is the router, and
        dispatch notification for the agent.
        """
        admin_context = context if context.is_admin else context.elevated()
        if not version:
            version = self.target.version

        vpn_agents = self.driver.service_plugin.get_vpn_agents_hosting_routers(
            admin_context, [router_id], active=True)

        for vpn_agent in vpn_agents:
            LOG.debug('Notify agent at %(topic)s.%(host)s the message '
                      '%(method)s %(args)s',
                      {'topic': self.topic,
                       'host': vpn_agent['host'],
                       'method': method,
                       'args': kwargs})
            cctxt = self.client.prepare(server=vpn_agent['host'],
                                        version=version)
            cctxt.cast(context, method, **kwargs)


class IPsecOvnVPNDriver(BaseOvnIPsecVPNDriver):
    """VPN Service Driver class for IPsec."""

    def create_rpc_conn(self):
        self.agent_rpc = IPsecOvnVpnAgentApi(
            topics.IPSEC_AGENT_TOPIC, BASE_IPSEC_VERSION, self)

    def start_rpc_listeners(self):
        self.endpoints = [IPsecVpnOvnDriverCallBack(self)]
        self.conn = n_rpc.Connection()
        self.conn.create_consumer(
            topics.IPSEC_DRIVER_TOPIC, self.endpoints, fanout=False)
        return self.conn.consume_in_threads()
